# Copyright (C) 2006-2010 Søren Roug, European Environment Agency
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Contributor(s):
#

import re

from .namespaces import (
    ANIMNS,
    CHARTNS,
    CONFIGNS,
    DR3DNS,
    DRAWNS,
    FONS,
    FORMNS,
    MANIFESTNS,
    METANS,
    NUMBERNS,
    OFFICENS,
    PRESENTATIONNS,
    SCRIPTNS,
    SMILNS,
    STYLENS,
    SVGNS,
    TABLENS,
    TEXTNS,
    XFORMSNS,
    XLINKNS,
)

pattern_color = re.compile(r'#[0-9a-fA-F]{6}')
pattern_vector3D = re.compile(r'\([ ]*-?([0-9]+(\.[0-9]*)?|\.[0-9]+)([ ]+-?([0-9]+(\.[0-9]*)?|\.[0-9]+)){2}[ ]*\)')


def make_NCName(arg):
    for c in (':', ' '):
        arg = arg.replace(c, f'_{ord(c):x}_')
    return arg


def cnv_anyURI(attribute, arg, element):
    return str(arg)


def cnv_boolean(attribute, arg, element):
    if arg.lower() in ('false', 'no'):
        return 'false'
    if arg:
        return 'true'
    return 'false'


# Potentially accept color values


def cnv_color(attribute, arg, element):
    """A RGB color in conformance with §5.9.11 of [XSL], that is a RGB color in notation “#rrggbb”, where
    rr, gg and bb are 8-bit hexadecimal digits.
    """
    return str(arg)


def cnv_configtype(attribute, arg, element):
    if str(arg) not in ('boolean', 'short', 'int', 'long', 'double', 'string', 'datetime', 'base64Binary'):
        raise ValueError(f"'{arg!s}' not allowed")
    return str(arg)


def cnv_data_source_has_labels(attribute, arg, element):
    if str(arg) not in ('none', 'row', 'column', 'both'):
        raise ValueError(f"'{arg!s}' not allowed")
    return str(arg)


# Understand different date formats


def cnv_date(attribute, arg, element):
    """A dateOrDateTime value is either an [xmlschema-2] date value or an [xmlschema-2] dateTime
    value.
    """
    return str(arg)


def cnv_dateTime(attribute, arg, element):
    """A dateOrDateTime value is either an [xmlschema-2] date value or an [xmlschema-2] dateTime
    value.
    """
    return str(arg)


def cnv_double(attribute, arg, element):
    return str(arg)


def cnv_duration(attribute, arg, element):
    return str(arg)


def cnv_family(attribute, arg, element):
    """A style family"""
    if str(arg) not in (
        'text',
        'paragraph',
        'section',
        'ruby',
        'table',
        'table-column',
        'table-row',
        'table-cell',
        'graphic',
        'presentation',
        'drawing-page',
        'chart',
    ):
        raise ValueError(f"'{arg!s}' not allowed")
    return str(arg)


def __save_prefix(attribute, arg, element):
    prefix = arg.split(':', 1)[0]
    if prefix == arg:
        return str(arg)
    namespace = element.get_knownns(prefix)
    if namespace is None:
        # raise ValueError(f"'{str(prefix)}' is an unknown prefix")
        return str(arg)
    return str(arg)


def cnv_formula(attribute, arg, element):
    """A string containing a formula. Formulas do not have a predefined syntax, but the string should
    begin with a namespace prefix, followed by a “:” (COLON, U+003A) separator, followed by the text
    of the formula. The namespace bound to the prefix determines the syntax and semantics of the
    formula.
    """
    return __save_prefix(attribute, arg, element)


def cnv_ID(attribute, arg, element):
    return str(arg)


def cnv_IDREF(attribute, arg, element):
    return str(arg)


def cnv_integer(attribute, arg, element):
    return str(arg)


def cnv_legend_position(attribute, arg, element):
    if str(arg) not in ('start', 'end', 'top', 'bottom', 'top-start', 'bottom-start', 'top-end', 'bottom-end'):
        raise ValueError(f"'{arg!s}' not allowed")
    return str(arg)


pattern_length = re.compile(r'-?([0-9]+(\.[0-9]*)?|\.[0-9]+)((cm)|(mm)|(in)|(pt)|(pc)|(px))')


def cnv_length(attribute, arg, element):
    """A (positive or negative) physical length, consisting of magnitude and unit, in conformance with the
    Units of Measure defined in §5.9.13 of [XSL].
    """
    global pattern_length
    if not pattern_length.match(arg):
        raise ValueError(f"'{arg}' is not a valid length")
    return arg


def cnv_lengthorpercent(attribute, arg, element):
    failed = False
    try:
        return cnv_length(attribute, arg, element)
    except Exception:
        failed = True
    try:
        return cnv_percent(attribute, arg, element)
    except Exception:
        failed = True
    if failed:
        raise ValueError(f"'{arg}' is not a valid length or percent")
    return arg


def cnv_metavaluetype(attribute, arg, element):
    if str(arg) not in ('float', 'date', 'time', 'boolean', 'string'):
        raise ValueError(f"'{arg!s}' not allowed")
    return str(arg)


def cnv_major_minor(attribute, arg, element):
    if arg not in ('major', 'minor'):
        raise ValueError(f"'{arg}' is not either 'minor' or 'major'")


pattern_namespacedToken = re.compile(r'[0-9a-zA-Z_]+:[0-9a-zA-Z._\-]+')


def cnv_namespacedToken(attribute, arg, element):
    global pattern_namespacedToken

    if not pattern_namespacedToken.match(arg):
        raise ValueError(f"'{arg}' is not a valid namespaced token")
    return __save_prefix(attribute, arg, element)


def cnv_NCName(attribute, arg, element):
    """NCName is defined in http://www.w3.org/TR/REC-xml-names/#NT-NCName
    Essentially an XML name minus ':'
    """
    if isinstance(arg, (str, bytes)):
        return make_NCName(arg)
    else:
        return arg.getAttrNS(STYLENS, 'name')


# This function takes either an instance of a style (preferred)
# or a text string naming the style. If it is a text string, then it must
# already have been converted to an NCName
# The text-string argument is mainly for when we build a structure from XML


def cnv_StyleNameRef(attribute, arg, element):
    try:
        return arg.getAttrNS(STYLENS, 'name')
    except Exception:
        return arg


# This function takes either an instance of a style (preferred)
# or a text string naming the style. If it is a text string, then it must
# already have been converted to an NCName
# The text-string argument is mainly for when we build a structure from XML


def cnv_DrawNameRef(attribute, arg, element):
    try:
        return arg.getAttrNS(DRAWNS, 'name')
    except Exception:
        return arg


# Must accept list of Style objects


def cnv_NCNames(attribute, arg, element):
    return ' '.join(arg)


def cnv_nonNegativeInteger(attribute, arg, element):
    return str(arg)


pattern_percent = re.compile(r'-?([0-9]+(\.[0-9]*)?|\.[0-9]+)%')


def cnv_percent(attribute, arg, element):
    global pattern_percent
    if not pattern_percent.match(arg):
        raise ValueError(f"'{arg}' is not a valid length")
    return arg


# Real one doesn't allow floating point values
pattern_points = re.compile(r'-?[0-9]+,-?[0-9]+([ ]+-?[0-9]+,-?[0-9]+)*')
# pattern_points = re.compile(r'-?[0-9.]+,-?[0-9.]+([ ]+-?[0-9.]+,-?[0-9.]+)*')


def cnv_points(attribute, arg, element):
    global pattern_points
    if isinstance(arg, (str, bytes)):
        if not pattern_points.match(arg):
            raise ValueError('x,y are separated by a comma and the points are separated by white spaces')
        return arg
    else:
        try:
            strarg = ' '.join(['{},{}'.format(*p) for p in arg])
        except Exception:
            raise ValueError(f'Points must be string or [(0,0),(1,1)] - not {arg}')
        return strarg


def cnv_positiveInteger(attribute, arg, element):
    return str(arg)


def cnv_string(attribute, arg, element):
    return str(arg)


def cnv_textnoteclass(attribute, arg, element):
    if str(arg) not in ('footnote', 'endnote'):
        raise ValueError(f"'{arg!s}' not allowed")
    return str(arg)


# Understand different time formats


def cnv_time(attribute, arg, element):
    return str(arg)


def cnv_token(attribute, arg, element):
    return str(arg)


pattern_viewbox = re.compile(r'-?[0-9]+([ ]+-?[0-9]+){3}$')


def cnv_viewbox(attribute, arg, element):
    global pattern_viewbox
    if not pattern_viewbox.match(arg):
        raise ValueError('viewBox must be four integers separated by whitespaces')
    return arg


def cnv_xlinkshow(attribute, arg, element):
    if str(arg) not in ('new', 'replace', 'embed'):
        raise ValueError(f"'{arg!s}' not allowed")
    return str(arg)


attrconverters = {
    ((ANIMNS, 'audio-level'), None): cnv_double,
    ((ANIMNS, 'color-interpolation'), None): cnv_string,
    ((ANIMNS, 'color-interpolation-direction'), None): cnv_string,
    ((ANIMNS, 'command'), None): cnv_string,
    ((ANIMNS, 'formula'), None): cnv_string,
    ((ANIMNS, 'id'), None): cnv_ID,
    ((ANIMNS, 'iterate-interval'), None): cnv_duration,
    ((ANIMNS, 'iterate-type'), None): cnv_string,
    ((ANIMNS, 'name'), None): cnv_string,
    ((ANIMNS, 'sub-item'), None): cnv_string,
    ((ANIMNS, 'value'), None): cnv_string,
    # ((DBNS,'type'), None): cnv_namespacedToken,
    ((CHARTNS, 'attached-axis'), None): cnv_string,
    ((CHARTNS, 'class'), (CHARTNS, 'grid')): cnv_major_minor,
    ((CHARTNS, 'class'), None): cnv_namespacedToken,
    ((CHARTNS, 'column-mapping'), None): cnv_string,
    ((CHARTNS, 'connect-bars'), None): cnv_boolean,
    ((CHARTNS, 'data-label-number'), None): cnv_string,
    ((CHARTNS, 'data-label-symbol'), None): cnv_boolean,
    ((CHARTNS, 'data-label-text'), None): cnv_boolean,
    ((CHARTNS, 'data-source-has-labels'), None): cnv_data_source_has_labels,
    ((CHARTNS, 'deep'), None): cnv_boolean,
    ((CHARTNS, 'dimension'), None): cnv_string,
    ((CHARTNS, 'display-label'), None): cnv_boolean,
    ((CHARTNS, 'error-category'), None): cnv_string,
    ((CHARTNS, 'error-lower-indicator'), None): cnv_boolean,
    ((CHARTNS, 'error-lower-limit'), None): cnv_string,
    ((CHARTNS, 'error-margin'), None): cnv_string,
    ((CHARTNS, 'error-percentage'), None): cnv_string,
    ((CHARTNS, 'error-upper-indicator'), None): cnv_boolean,
    ((CHARTNS, 'error-upper-limit'), None): cnv_string,
    ((CHARTNS, 'gap-width'), None): cnv_string,
    ((CHARTNS, 'interpolation'), None): cnv_string,
    ((CHARTNS, 'interval-major'), None): cnv_string,
    ((CHARTNS, 'interval-minor-divisor'), None): cnv_string,
    ((CHARTNS, 'japanese-candle-stick'), None): cnv_boolean,
    ((CHARTNS, 'label-arrangement'), None): cnv_string,
    ((CHARTNS, 'label-cell-address'), None): cnv_string,
    ((CHARTNS, 'legend-align'), None): cnv_string,
    ((CHARTNS, 'legend-position'), None): cnv_legend_position,
    ((CHARTNS, 'lines'), None): cnv_boolean,
    ((CHARTNS, 'link-data-style-to-source'), None): cnv_boolean,
    ((CHARTNS, 'logarithmic'), None): cnv_boolean,
    ((CHARTNS, 'maximum'), None): cnv_string,
    ((CHARTNS, 'mean-value'), None): cnv_boolean,
    ((CHARTNS, 'minimum'), None): cnv_string,
    ((CHARTNS, 'name'), None): cnv_string,
    ((CHARTNS, 'origin'), None): cnv_string,
    ((CHARTNS, 'overlap'), None): cnv_string,
    ((CHARTNS, 'percentage'), None): cnv_boolean,
    ((CHARTNS, 'pie-offset'), None): cnv_string,
    ((CHARTNS, 'regression-type'), None): cnv_string,
    ((CHARTNS, 'repeated'), None): cnv_nonNegativeInteger,
    ((CHARTNS, 'row-mapping'), None): cnv_string,
    ((CHARTNS, 'scale-text'), None): cnv_boolean,
    ((CHARTNS, 'series-source'), None): cnv_string,
    ((CHARTNS, 'solid-type'), None): cnv_string,
    ((CHARTNS, 'spline-order'), None): cnv_string,
    ((CHARTNS, 'spline-resolution'), None): cnv_string,
    ((CHARTNS, 'stacked'), None): cnv_boolean,
    ((CHARTNS, 'style-name'), None): cnv_StyleNameRef,
    ((CHARTNS, 'symbol-height'), None): cnv_string,
    ((CHARTNS, 'symbol-name'), None): cnv_string,
    ((CHARTNS, 'symbol-type'), None): cnv_string,
    ((CHARTNS, 'symbol-width'), None): cnv_string,
    ((CHARTNS, 'text-overlap'), None): cnv_boolean,
    ((CHARTNS, 'three-dimensional'), None): cnv_boolean,
    ((CHARTNS, 'tick-marks-major-inner'), None): cnv_boolean,
    ((CHARTNS, 'tick-marks-major-outer'), None): cnv_boolean,
    ((CHARTNS, 'tick-marks-minor-inner'), None): cnv_boolean,
    ((CHARTNS, 'tick-marks-minor-outer'), None): cnv_boolean,
    ((CHARTNS, 'values-cell-range-address'), None): cnv_string,
    ((CHARTNS, 'vertical'), None): cnv_boolean,
    ((CHARTNS, 'visible'), None): cnv_boolean,
    ((CONFIGNS, 'name'), None): cnv_formula,
    ((CONFIGNS, 'type'), None): cnv_configtype,
    ((DR3DNS, 'ambient-color'), None): cnv_string,
    ((DR3DNS, 'back-scale'), None): cnv_string,
    ((DR3DNS, 'backface-culling'), None): cnv_string,
    ((DR3DNS, 'center'), None): cnv_string,
    ((DR3DNS, 'close-back'), None): cnv_boolean,
    ((DR3DNS, 'close-front'), None): cnv_boolean,
    ((DR3DNS, 'depth'), None): cnv_length,
    ((DR3DNS, 'diffuse-color'), None): cnv_string,
    ((DR3DNS, 'direction'), None): cnv_string,
    ((DR3DNS, 'distance'), None): cnv_length,
    ((DR3DNS, 'edge-rounding'), None): cnv_string,
    ((DR3DNS, 'edge-rounding-mode'), None): cnv_string,
    ((DR3DNS, 'emissive-color'), None): cnv_string,
    ((DR3DNS, 'enabled'), None): cnv_boolean,
    ((DR3DNS, 'end-angle'), None): cnv_string,
    ((DR3DNS, 'focal-length'), None): cnv_length,
    ((DR3DNS, 'horizontal-segments'), None): cnv_string,
    ((DR3DNS, 'lighting-mode'), None): cnv_boolean,
    ((DR3DNS, 'max-edge'), None): cnv_string,
    ((DR3DNS, 'min-edge'), None): cnv_string,
    ((DR3DNS, 'normals-direction'), None): cnv_string,
    ((DR3DNS, 'normals-kind'), None): cnv_string,
    ((DR3DNS, 'projection'), None): cnv_string,
    ((DR3DNS, 'shade-mode'), None): cnv_string,
    ((DR3DNS, 'shadow'), None): cnv_string,
    ((DR3DNS, 'shadow-slant'), None): cnv_nonNegativeInteger,
    ((DR3DNS, 'shininess'), None): cnv_string,
    ((DR3DNS, 'size'), None): cnv_string,
    ((DR3DNS, 'specular'), None): cnv_boolean,
    ((DR3DNS, 'specular-color'), None): cnv_string,
    ((DR3DNS, 'texture-filter'), None): cnv_string,
    ((DR3DNS, 'texture-generation-mode-x'), None): cnv_string,
    ((DR3DNS, 'texture-generation-mode-y'), None): cnv_string,
    ((DR3DNS, 'texture-kind'), None): cnv_string,
    ((DR3DNS, 'texture-mode'), None): cnv_string,
    ((DR3DNS, 'transform'), None): cnv_string,
    ((DR3DNS, 'vertical-segments'), None): cnv_string,
    ((DR3DNS, 'vpn'), None): cnv_string,
    ((DR3DNS, 'vrp'), None): cnv_string,
    ((DR3DNS, 'vup'), None): cnv_string,
    ((DRAWNS, 'align'), None): cnv_string,
    ((DRAWNS, 'angle'), None): cnv_integer,
    ((DRAWNS, 'archive'), None): cnv_string,
    ((DRAWNS, 'auto-grow-height'), None): cnv_boolean,
    ((DRAWNS, 'auto-grow-width'), None): cnv_boolean,
    ((DRAWNS, 'background-size'), None): cnv_string,
    ((DRAWNS, 'blue'), None): cnv_string,
    ((DRAWNS, 'border'), None): cnv_string,
    ((DRAWNS, 'caption-angle'), None): cnv_string,
    ((DRAWNS, 'caption-angle-type'), None): cnv_string,
    ((DRAWNS, 'caption-escape'), None): cnv_string,
    ((DRAWNS, 'caption-escape-direction'), None): cnv_string,
    ((DRAWNS, 'caption-fit-line-length'), None): cnv_boolean,
    ((DRAWNS, 'caption-gap'), None): cnv_string,
    ((DRAWNS, 'caption-line-length'), None): cnv_length,
    ((DRAWNS, 'caption-point-x'), None): cnv_string,
    ((DRAWNS, 'caption-point-y'), None): cnv_string,
    ((DRAWNS, 'caption-id'), None): cnv_IDREF,
    ((DRAWNS, 'caption-type'), None): cnv_string,
    ((DRAWNS, 'chain-next-name'), None): cnv_string,
    ((DRAWNS, 'class-id'), None): cnv_string,
    ((DRAWNS, 'class-names'), None): cnv_NCNames,
    ((DRAWNS, 'code'), None): cnv_string,
    ((DRAWNS, 'color'), None): cnv_string,
    ((DRAWNS, 'color-inversion'), None): cnv_boolean,
    ((DRAWNS, 'color-mode'), None): cnv_string,
    ((DRAWNS, 'concave'), None): cnv_string,
    ((DRAWNS, 'concentric-gradient-fill-allowed'), None): cnv_boolean,
    ((DRAWNS, 'contrast'), None): cnv_string,
    ((DRAWNS, 'control'), None): cnv_IDREF,
    ((DRAWNS, 'copy-of'), None): cnv_string,
    ((DRAWNS, 'corner-radius'), None): cnv_length,
    ((DRAWNS, 'corners'), None): cnv_positiveInteger,
    ((DRAWNS, 'cx'), None): cnv_string,
    ((DRAWNS, 'cy'), None): cnv_string,
    ((DRAWNS, 'data'), None): cnv_string,
    ((DRAWNS, 'decimal-places'), None): cnv_string,
    ((DRAWNS, 'display'), None): cnv_string,
    ((DRAWNS, 'display-name'), None): cnv_string,
    ((DRAWNS, 'distance'), None): cnv_lengthorpercent,
    ((DRAWNS, 'dots1'), None): cnv_integer,
    ((DRAWNS, 'dots1-length'), None): cnv_lengthorpercent,
    ((DRAWNS, 'dots2'), None): cnv_integer,
    ((DRAWNS, 'dots2-length'), None): cnv_lengthorpercent,
    ((DRAWNS, 'end-angle'), None): cnv_double,
    ((DRAWNS, 'end'), None): cnv_string,
    ((DRAWNS, 'end-color'), None): cnv_string,
    ((DRAWNS, 'end-glue-point'), None): cnv_nonNegativeInteger,
    ((DRAWNS, 'end-guide'), None): cnv_length,
    ((DRAWNS, 'end-intensity'), None): cnv_string,
    ((DRAWNS, 'end-line-spacing-horizontal'), None): cnv_string,
    ((DRAWNS, 'end-line-spacing-vertical'), None): cnv_string,
    ((DRAWNS, 'end-shape'), None): cnv_IDREF,
    ((DRAWNS, 'engine'), None): cnv_namespacedToken,
    ((DRAWNS, 'enhanced-path'), None): cnv_string,
    ((DRAWNS, 'escape-direction'), None): cnv_string,
    ((DRAWNS, 'extrusion-allowed'), None): cnv_boolean,
    ((DRAWNS, 'extrusion-brightness'), None): cnv_string,
    ((DRAWNS, 'extrusion'), None): cnv_boolean,
    ((DRAWNS, 'extrusion-color'), None): cnv_boolean,
    ((DRAWNS, 'extrusion-depth'), None): cnv_double,
    ((DRAWNS, 'extrusion-diffusion'), None): cnv_string,
    ((DRAWNS, 'extrusion-first-light-direction'), None): cnv_string,
    ((DRAWNS, 'extrusion-first-light-harsh'), None): cnv_boolean,
    ((DRAWNS, 'extrusion-first-light-level'), None): cnv_string,
    ((DRAWNS, 'extrusion-light-face'), None): cnv_boolean,
    ((DRAWNS, 'extrusion-metal'), None): cnv_boolean,
    ((DRAWNS, 'extrusion-number-of-line-segments'), None): cnv_integer,
    ((DRAWNS, 'extrusion-origin'), None): cnv_double,
    ((DRAWNS, 'extrusion-rotation-angle'), None): cnv_double,
    ((DRAWNS, 'extrusion-rotation-center'), None): cnv_string,
    ((DRAWNS, 'extrusion-second-light-direction'), None): cnv_string,
    ((DRAWNS, 'extrusion-second-light-harsh'), None): cnv_boolean,
    ((DRAWNS, 'extrusion-second-light-level'), None): cnv_string,
    ((DRAWNS, 'extrusion-shininess'), None): cnv_string,
    ((DRAWNS, 'extrusion-skew'), None): cnv_double,
    ((DRAWNS, 'extrusion-specularity'), None): cnv_string,
    ((DRAWNS, 'extrusion-viewpoint'), None): cnv_string,
    ((DRAWNS, 'fill'), None): cnv_string,
    ((DRAWNS, 'fill-color'), None): cnv_string,
    ((DRAWNS, 'fill-gradient-name'), None): cnv_string,
    ((DRAWNS, 'fill-hatch-name'), None): cnv_string,
    ((DRAWNS, 'fill-hatch-solid'), None): cnv_boolean,
    ((DRAWNS, 'fill-image-height'), None): cnv_lengthorpercent,
    ((DRAWNS, 'fill-image-name'), None): cnv_DrawNameRef,
    ((DRAWNS, 'fill-image-ref-point'), None): cnv_string,
    ((DRAWNS, 'fill-image-ref-point-x'), None): cnv_string,
    ((DRAWNS, 'fill-image-ref-point-y'), None): cnv_string,
    ((DRAWNS, 'fill-image-width'), None): cnv_lengthorpercent,
    ((DRAWNS, 'filter-name'), None): cnv_string,
    ((DRAWNS, 'fit-to-contour'), None): cnv_boolean,
    ((DRAWNS, 'fit-to-size'), None): cnv_boolean,
    ((DRAWNS, 'formula'), None): cnv_string,
    ((DRAWNS, 'frame-display-border'), None): cnv_boolean,
    ((DRAWNS, 'frame-display-scrollbar'), None): cnv_boolean,
    ((DRAWNS, 'frame-margin-horizontal'), None): cnv_string,
    ((DRAWNS, 'frame-margin-vertical'), None): cnv_string,
    ((DRAWNS, 'frame-name'), None): cnv_string,
    ((DRAWNS, 'gamma'), None): cnv_string,
    ((DRAWNS, 'glue-point-leaving-directions'), None): cnv_string,
    ((DRAWNS, 'glue-point-type'), None): cnv_string,
    ((DRAWNS, 'glue-points'), None): cnv_string,
    ((DRAWNS, 'gradient-step-count'), None): cnv_string,
    ((DRAWNS, 'green'), None): cnv_string,
    ((DRAWNS, 'guide-distance'), None): cnv_string,
    ((DRAWNS, 'guide-overhang'), None): cnv_length,
    ((DRAWNS, 'handle-mirror-horizontal'), None): cnv_boolean,
    ((DRAWNS, 'handle-mirror-vertical'), None): cnv_boolean,
    ((DRAWNS, 'handle-polar'), None): cnv_string,
    ((DRAWNS, 'handle-position'), None): cnv_string,
    ((DRAWNS, 'handle-radius-range-maximum'), None): cnv_string,
    ((DRAWNS, 'handle-radius-range-minimum'), None): cnv_string,
    ((DRAWNS, 'handle-range-x-maximum'), None): cnv_string,
    ((DRAWNS, 'handle-range-x-minimum'), None): cnv_string,
    ((DRAWNS, 'handle-range-y-maximum'), None): cnv_string,
    ((DRAWNS, 'handle-range-y-minimum'), None): cnv_string,
    ((DRAWNS, 'handle-switched'), None): cnv_boolean,
    # ((DRAWNS,'id'), None): cnv_ID,
    # ((DRAWNS,'id'), None): cnv_nonNegativeInteger,   # ?? line 6581 in RNG
    ((DRAWNS, 'id'), None): cnv_string,
    ((DRAWNS, 'image-opacity'), None): cnv_string,
    ((DRAWNS, 'kind'), None): cnv_string,
    ((DRAWNS, 'layer'), None): cnv_string,
    ((DRAWNS, 'line-distance'), None): cnv_string,
    ((DRAWNS, 'line-skew'), None): cnv_string,
    ((DRAWNS, 'luminance'), None): cnv_string,
    ((DRAWNS, 'marker-end-center'), None): cnv_boolean,
    ((DRAWNS, 'marker-end'), None): cnv_string,
    ((DRAWNS, 'marker-end-width'), None): cnv_length,
    ((DRAWNS, 'marker-start-center'), None): cnv_boolean,
    ((DRAWNS, 'marker-start'), None): cnv_string,
    ((DRAWNS, 'marker-start-width'), None): cnv_length,
    ((DRAWNS, 'master-page-name'), None): cnv_StyleNameRef,
    ((DRAWNS, 'may-script'), None): cnv_boolean,
    ((DRAWNS, 'measure-align'), None): cnv_string,
    ((DRAWNS, 'measure-vertical-align'), None): cnv_string,
    ((DRAWNS, 'mime-type'), None): cnv_string,
    ((DRAWNS, 'mirror-horizontal'), None): cnv_boolean,
    ((DRAWNS, 'mirror-vertical'), None): cnv_boolean,
    ((DRAWNS, 'modifiers'), None): cnv_string,
    ((DRAWNS, 'name'), None): cnv_NCName,
    # ((DRAWNS,'name'), None): cnv_string,
    ((DRAWNS, 'nav-order'), None): cnv_IDREF,
    ((DRAWNS, 'nohref'), None): cnv_string,
    ((DRAWNS, 'notify-on-update-of-ranges'), None): cnv_string,
    ((DRAWNS, 'object'), None): cnv_string,
    ((DRAWNS, 'ole-draw-aspect'), None): cnv_string,
    ((DRAWNS, 'opacity'), None): cnv_string,
    ((DRAWNS, 'opacity-name'), None): cnv_string,
    ((DRAWNS, 'page-number'), None): cnv_positiveInteger,
    ((DRAWNS, 'parallel'), None): cnv_boolean,
    ((DRAWNS, 'path-stretchpoint-x'), None): cnv_double,
    ((DRAWNS, 'path-stretchpoint-y'), None): cnv_double,
    ((DRAWNS, 'placing'), None): cnv_string,
    ((DRAWNS, 'points'), None): cnv_points,
    ((DRAWNS, 'protected'), None): cnv_boolean,
    ((DRAWNS, 'recreate-on-edit'), None): cnv_boolean,
    ((DRAWNS, 'red'), None): cnv_string,
    ((DRAWNS, 'rotation'), None): cnv_integer,
    ((DRAWNS, 'secondary-fill-color'), None): cnv_string,
    ((DRAWNS, 'shadow'), None): cnv_string,
    ((DRAWNS, 'shadow-color'), None): cnv_string,
    ((DRAWNS, 'shadow-offset-x'), None): cnv_length,
    ((DRAWNS, 'shadow-offset-y'), None): cnv_length,
    ((DRAWNS, 'shadow-opacity'), None): cnv_string,
    ((DRAWNS, 'shape-id'), None): cnv_IDREF,
    ((DRAWNS, 'sharpness'), None): cnv_string,
    ((DRAWNS, 'show-unit'), None): cnv_boolean,
    ((DRAWNS, 'start-angle'), None): cnv_double,
    ((DRAWNS, 'start'), None): cnv_string,
    ((DRAWNS, 'start-color'), None): cnv_string,
    ((DRAWNS, 'start-glue-point'), None): cnv_nonNegativeInteger,
    ((DRAWNS, 'start-guide'), None): cnv_length,
    ((DRAWNS, 'start-intensity'), None): cnv_string,
    ((DRAWNS, 'start-line-spacing-horizontal'), None): cnv_string,
    ((DRAWNS, 'start-line-spacing-vertical'), None): cnv_string,
    ((DRAWNS, 'start-shape'), None): cnv_IDREF,
    ((DRAWNS, 'stroke'), None): cnv_string,
    ((DRAWNS, 'stroke-dash'), None): cnv_string,
    ((DRAWNS, 'stroke-dash-names'), None): cnv_string,
    ((DRAWNS, 'stroke-linejoin'), None): cnv_string,
    ((DRAWNS, 'style'), None): cnv_string,
    ((DRAWNS, 'style-name'), None): cnv_StyleNameRef,
    ((DRAWNS, 'symbol-color'), None): cnv_string,
    ((DRAWNS, 'text-areas'), None): cnv_string,
    ((DRAWNS, 'text-path-allowed'), None): cnv_boolean,
    ((DRAWNS, 'text-path'), None): cnv_boolean,
    ((DRAWNS, 'text-path-mode'), None): cnv_string,
    ((DRAWNS, 'text-path-same-letter-heights'), None): cnv_boolean,
    ((DRAWNS, 'text-path-scale'), None): cnv_string,
    ((DRAWNS, 'text-rotate-angle'), None): cnv_double,
    ((DRAWNS, 'text-style-name'), None): cnv_StyleNameRef,
    ((DRAWNS, 'textarea-horizontal-align'), None): cnv_string,
    ((DRAWNS, 'textarea-vertical-align'), None): cnv_string,
    ((DRAWNS, 'tile-repeat-offset'), None): cnv_string,
    ((DRAWNS, 'transform'), None): cnv_string,
    ((DRAWNS, 'type'), None): cnv_string,
    ((DRAWNS, 'unit'), None): cnv_string,
    ((DRAWNS, 'value'), None): cnv_string,
    ((DRAWNS, 'visible-area-height'), None): cnv_string,
    ((DRAWNS, 'visible-area-left'), None): cnv_string,
    ((DRAWNS, 'visible-area-top'), None): cnv_string,
    ((DRAWNS, 'visible-area-width'), None): cnv_string,
    ((DRAWNS, 'wrap-influence-on-position'), None): cnv_string,
    ((DRAWNS, 'z-index'), None): cnv_nonNegativeInteger,
    ((FONS, 'background-color'), None): cnv_string,
    ((FONS, 'border-bottom'), None): cnv_string,
    ((FONS, 'border'), None): cnv_string,
    ((FONS, 'border-left'), None): cnv_string,
    ((FONS, 'border-right'), None): cnv_string,
    ((FONS, 'border-top'), None): cnv_string,
    ((FONS, 'break-after'), None): cnv_string,
    ((FONS, 'break-before'), None): cnv_string,
    ((FONS, 'clip'), None): cnv_string,
    ((FONS, 'color'), None): cnv_string,
    ((FONS, 'column-count'), None): cnv_positiveInteger,
    ((FONS, 'column-gap'), None): cnv_length,
    ((FONS, 'country'), None): cnv_token,
    ((FONS, 'end-indent'), None): cnv_length,
    ((FONS, 'font-family'), None): cnv_string,
    ((FONS, 'font-size'), None): cnv_string,
    ((FONS, 'font-style'), None): cnv_string,
    ((FONS, 'font-variant'), None): cnv_string,
    ((FONS, 'font-weight'), None): cnv_string,
    ((FONS, 'height'), None): cnv_string,
    ((FONS, 'hyphenate'), None): cnv_boolean,
    ((FONS, 'hyphenation-keep'), None): cnv_string,
    ((FONS, 'hyphenation-ladder-count'), None): cnv_string,
    ((FONS, 'hyphenation-push-char-count'), None): cnv_string,
    ((FONS, 'hyphenation-remain-char-count'), None): cnv_string,
    ((FONS, 'keep-together'), None): cnv_string,
    ((FONS, 'keep-with-next'), None): cnv_string,
    ((FONS, 'language'), None): cnv_token,
    ((FONS, 'letter-spacing'), None): cnv_string,
    ((FONS, 'line-height'), None): cnv_string,
    ((FONS, 'margin-bottom'), None): cnv_string,
    ((FONS, 'margin'), None): cnv_string,
    ((FONS, 'margin-left'), None): cnv_string,
    ((FONS, 'margin-right'), None): cnv_string,
    ((FONS, 'margin-top'), None): cnv_string,
    ((FONS, 'max-height'), None): cnv_string,
    ((FONS, 'max-width'), None): cnv_string,
    ((FONS, 'min-height'), None): cnv_lengthorpercent,
    ((FONS, 'min-width'), None): cnv_string,
    ((FONS, 'orphans'), None): cnv_string,
    ((FONS, 'padding-bottom'), None): cnv_string,
    ((FONS, 'padding'), None): cnv_string,
    ((FONS, 'padding-left'), None): cnv_string,
    ((FONS, 'padding-right'), None): cnv_string,
    ((FONS, 'padding-top'), None): cnv_string,
    ((FONS, 'page-height'), None): cnv_length,
    ((FONS, 'page-width'), None): cnv_length,
    ((FONS, 'space-after'), None): cnv_length,
    ((FONS, 'space-before'), None): cnv_length,
    ((FONS, 'start-indent'), None): cnv_length,
    ((FONS, 'text-align'), None): cnv_string,
    ((FONS, 'text-align-last'), None): cnv_string,
    ((FONS, 'text-indent'), None): cnv_string,
    ((FONS, 'text-shadow'), None): cnv_string,
    ((FONS, 'text-transform'), None): cnv_string,
    ((FONS, 'widows'), None): cnv_string,
    ((FONS, 'width'), None): cnv_string,
    ((FONS, 'wrap-option'), None): cnv_string,
    ((FORMNS, 'allow-deletes'), None): cnv_boolean,
    ((FORMNS, 'allow-inserts'), None): cnv_boolean,
    ((FORMNS, 'allow-updates'), None): cnv_boolean,
    ((FORMNS, 'apply-design-mode'), None): cnv_boolean,
    ((FORMNS, 'apply-filter'), None): cnv_boolean,
    ((FORMNS, 'auto-complete'), None): cnv_boolean,
    ((FORMNS, 'automatic-focus'), None): cnv_boolean,
    ((FORMNS, 'bound-column'), None): cnv_string,
    ((FORMNS, 'button-type'), None): cnv_string,
    ((FORMNS, 'command'), None): cnv_string,
    ((FORMNS, 'command-type'), None): cnv_string,
    ((FORMNS, 'control-implementation'), None): cnv_namespacedToken,
    ((FORMNS, 'convert-empty-to-null'), None): cnv_boolean,
    ((FORMNS, 'current-selected'), None): cnv_boolean,
    ((FORMNS, 'current-state'), None): cnv_string,
    # ((FORMNS,'current-value'), None): cnv_date,
    # ((FORMNS,'current-value'), None): cnv_double,
    ((FORMNS, 'current-value'), None): cnv_string,
    # ((FORMNS,'current-value'), None): cnv_time,
    ((FORMNS, 'data-field'), None): cnv_string,
    ((FORMNS, 'datasource'), None): cnv_string,
    ((FORMNS, 'default-button'), None): cnv_boolean,
    ((FORMNS, 'delay-for-repeat'), None): cnv_duration,
    ((FORMNS, 'detail-fields'), None): cnv_string,
    ((FORMNS, 'disabled'), None): cnv_boolean,
    ((FORMNS, 'dropdown'), None): cnv_boolean,
    ((FORMNS, 'echo-char'), None): cnv_string,
    ((FORMNS, 'enctype'), None): cnv_string,
    ((FORMNS, 'escape-processing'), None): cnv_boolean,
    ((FORMNS, 'filter'), None): cnv_string,
    ((FORMNS, 'focus-on-click'), None): cnv_boolean,
    ((FORMNS, 'for'), None): cnv_string,
    ((FORMNS, 'id'), None): cnv_ID,
    ((FORMNS, 'ignore-result'), None): cnv_boolean,
    ((FORMNS, 'image-align'), None): cnv_string,
    ((FORMNS, 'image-data'), None): cnv_anyURI,
    ((FORMNS, 'image-position'), None): cnv_string,
    ((FORMNS, 'is-tristate'), None): cnv_boolean,
    ((FORMNS, 'label'), None): cnv_string,
    ((FORMNS, 'list-source'), None): cnv_string,
    ((FORMNS, 'list-source-type'), None): cnv_string,
    ((FORMNS, 'master-fields'), None): cnv_string,
    ((FORMNS, 'max-length'), None): cnv_nonNegativeInteger,
    # ((FORMNS,'max-value'), None): cnv_date,
    # ((FORMNS,'max-value'), None): cnv_double,
    ((FORMNS, 'max-value'), None): cnv_string,
    # ((FORMNS,'max-value'), None): cnv_time,
    ((FORMNS, 'method'), None): cnv_string,
    # ((FORMNS,'min-value'), None): cnv_date,
    # ((FORMNS,'min-value'), None): cnv_double,
    ((FORMNS, 'min-value'), None): cnv_string,
    # ((FORMNS,'min-value'), None): cnv_time,
    ((FORMNS, 'multi-line'), None): cnv_boolean,
    ((FORMNS, 'multiple'), None): cnv_boolean,
    ((FORMNS, 'name'), None): cnv_string,
    ((FORMNS, 'navigation-mode'), None): cnv_string,
    ((FORMNS, 'order'), None): cnv_string,
    ((FORMNS, 'orientation'), None): cnv_string,
    ((FORMNS, 'page-step-size'), None): cnv_positiveInteger,
    ((FORMNS, 'printable'), None): cnv_boolean,
    ((FORMNS, 'property-name'), None): cnv_string,
    ((FORMNS, 'readonly'), None): cnv_boolean,
    ((FORMNS, 'selected'), None): cnv_boolean,
    ((FORMNS, 'size'), None): cnv_nonNegativeInteger,
    ((FORMNS, 'state'), None): cnv_string,
    ((FORMNS, 'step-size'), None): cnv_positiveInteger,
    ((FORMNS, 'tab-cycle'), None): cnv_string,
    ((FORMNS, 'tab-index'), None): cnv_nonNegativeInteger,
    ((FORMNS, 'tab-stop'), None): cnv_boolean,
    ((FORMNS, 'text-style-name'), None): cnv_StyleNameRef,
    ((FORMNS, 'title'), None): cnv_string,
    ((FORMNS, 'toggle'), None): cnv_boolean,
    ((FORMNS, 'validation'), None): cnv_boolean,
    # ((FORMNS,'value'), None): cnv_date,
    # ((FORMNS,'value'), None): cnv_double,
    ((FORMNS, 'value'), None): cnv_string,
    # ((FORMNS,'value'), None): cnv_time,
    ((FORMNS, 'visual-effect'), None): cnv_string,
    ((FORMNS, 'xforms-list-source'), None): cnv_string,
    ((FORMNS, 'xforms-submission'), None): cnv_string,
    ((MANIFESTNS, 'algorithm-name'), None): cnv_string,
    ((MANIFESTNS, 'checksum'), None): cnv_string,
    ((MANIFESTNS, 'checksum-type'), None): cnv_string,
    ((MANIFESTNS, 'full-path'), None): cnv_string,
    ((MANIFESTNS, 'initialisation-vector'), None): cnv_string,
    ((MANIFESTNS, 'iteration-count'), None): cnv_nonNegativeInteger,
    ((MANIFESTNS, 'key-derivation-name'), None): cnv_string,
    ((MANIFESTNS, 'media-type'), None): cnv_string,
    ((MANIFESTNS, 'salt'), None): cnv_string,
    ((MANIFESTNS, 'size'), None): cnv_nonNegativeInteger,
    ((METANS, 'cell-count'), None): cnv_nonNegativeInteger,
    ((METANS, 'character-count'), None): cnv_nonNegativeInteger,
    ((METANS, 'date'), None): cnv_dateTime,
    ((METANS, 'delay'), None): cnv_duration,
    ((METANS, 'draw-count'), None): cnv_nonNegativeInteger,
    ((METANS, 'frame-count'), None): cnv_nonNegativeInteger,
    ((METANS, 'image-count'), None): cnv_nonNegativeInteger,
    ((METANS, 'name'), None): cnv_string,
    ((METANS, 'non-whitespace-character-count'), None): cnv_nonNegativeInteger,
    ((METANS, 'object-count'), None): cnv_nonNegativeInteger,
    ((METANS, 'ole-object-count'), None): cnv_nonNegativeInteger,
    ((METANS, 'page-count'), None): cnv_nonNegativeInteger,
    ((METANS, 'paragraph-count'), None): cnv_nonNegativeInteger,
    ((METANS, 'row-count'), None): cnv_nonNegativeInteger,
    ((METANS, 'sentence-count'), None): cnv_nonNegativeInteger,
    ((METANS, 'syllable-count'), None): cnv_nonNegativeInteger,
    ((METANS, 'table-count'), None): cnv_nonNegativeInteger,
    ((METANS, 'value-type'), None): cnv_metavaluetype,
    ((METANS, 'word-count'), None): cnv_nonNegativeInteger,
    ((NUMBERNS, 'automatic-order'), None): cnv_boolean,
    ((NUMBERNS, 'calendar'), None): cnv_string,
    ((NUMBERNS, 'country'), None): cnv_token,
    ((NUMBERNS, 'decimal-places'), None): cnv_integer,
    ((NUMBERNS, 'decimal-replacement'), None): cnv_string,
    ((NUMBERNS, 'denominator-value'), None): cnv_integer,
    ((NUMBERNS, 'display-factor'), None): cnv_double,
    ((NUMBERNS, 'format-source'), None): cnv_string,
    ((NUMBERNS, 'grouping'), None): cnv_boolean,
    ((NUMBERNS, 'language'), None): cnv_token,
    ((NUMBERNS, 'min-denominator-digits'), None): cnv_integer,
    ((NUMBERNS, 'min-exponent-digits'), None): cnv_integer,
    ((NUMBERNS, 'min-integer-digits'), None): cnv_integer,
    ((NUMBERNS, 'min-numerator-digits'), None): cnv_integer,
    ((NUMBERNS, 'position'), None): cnv_integer,
    ((NUMBERNS, 'possessive-form'), None): cnv_boolean,
    ((NUMBERNS, 'style'), None): cnv_string,
    ((NUMBERNS, 'textual'), None): cnv_boolean,
    ((NUMBERNS, 'title'), None): cnv_string,
    ((NUMBERNS, 'transliteration-country'), None): cnv_token,
    ((NUMBERNS, 'transliteration-format'), None): cnv_string,
    ((NUMBERNS, 'transliteration-language'), None): cnv_token,
    ((NUMBERNS, 'transliteration-style'), None): cnv_string,
    ((NUMBERNS, 'truncate-on-overflow'), None): cnv_boolean,
    ((OFFICENS, 'automatic-update'), None): cnv_boolean,
    ((OFFICENS, 'boolean-value'), None): cnv_boolean,
    ((OFFICENS, 'conversion-mode'), None): cnv_string,
    ((OFFICENS, 'currency'), None): cnv_string,
    ((OFFICENS, 'date-value'), None): cnv_dateTime,
    ((OFFICENS, 'dde-application'), None): cnv_string,
    ((OFFICENS, 'dde-item'), None): cnv_string,
    ((OFFICENS, 'dde-topic'), None): cnv_string,
    ((OFFICENS, 'display'), None): cnv_boolean,
    ((OFFICENS, 'mimetype'), None): cnv_string,
    ((OFFICENS, 'name'), None): cnv_string,
    ((OFFICENS, 'process-content'), None): cnv_boolean,
    ((OFFICENS, 'server-map'), None): cnv_boolean,
    ((OFFICENS, 'string-value'), None): cnv_string,
    ((OFFICENS, 'target-frame'), None): cnv_string,
    ((OFFICENS, 'target-frame-name'), None): cnv_string,
    ((OFFICENS, 'time-value'), None): cnv_duration,
    ((OFFICENS, 'title'), None): cnv_string,
    ((OFFICENS, 'value'), None): cnv_double,
    ((OFFICENS, 'value-type'), None): cnv_string,
    ((OFFICENS, 'version'), None): cnv_string,
    ((PRESENTATIONNS, 'action'), None): cnv_string,
    ((PRESENTATIONNS, 'animations'), None): cnv_string,
    ((PRESENTATIONNS, 'background-objects-visible'), None): cnv_boolean,
    ((PRESENTATIONNS, 'background-visible'), None): cnv_boolean,
    ((PRESENTATIONNS, 'class'), None): cnv_string,
    ((PRESENTATIONNS, 'class-names'), None): cnv_NCNames,
    ((PRESENTATIONNS, 'delay'), None): cnv_duration,
    ((PRESENTATIONNS, 'direction'), None): cnv_string,
    ((PRESENTATIONNS, 'display-date-time'), None): cnv_boolean,
    ((PRESENTATIONNS, 'display-footer'), None): cnv_boolean,
    ((PRESENTATIONNS, 'display-header'), None): cnv_boolean,
    ((PRESENTATIONNS, 'display-page-number'), None): cnv_boolean,
    ((PRESENTATIONNS, 'duration'), None): cnv_string,
    ((PRESENTATIONNS, 'effect'), None): cnv_string,
    ((PRESENTATIONNS, 'endless'), None): cnv_boolean,
    ((PRESENTATIONNS, 'force-manual'), None): cnv_boolean,
    ((PRESENTATIONNS, 'full-screen'), None): cnv_boolean,
    ((PRESENTATIONNS, 'group-id'), None): cnv_string,
    ((PRESENTATIONNS, 'master-element'), None): cnv_IDREF,
    ((PRESENTATIONNS, 'mouse-as-pen'), None): cnv_boolean,
    ((PRESENTATIONNS, 'mouse-visible'), None): cnv_boolean,
    ((PRESENTATIONNS, 'name'), None): cnv_string,
    ((PRESENTATIONNS, 'node-type'), None): cnv_string,
    ((PRESENTATIONNS, 'object'), None): cnv_string,
    ((PRESENTATIONNS, 'pages'), None): cnv_string,
    ((PRESENTATIONNS, 'path-id'), None): cnv_string,
    ((PRESENTATIONNS, 'pause'), None): cnv_duration,
    ((PRESENTATIONNS, 'placeholder'), None): cnv_boolean,
    ((PRESENTATIONNS, 'play-full'), None): cnv_boolean,
    ((PRESENTATIONNS, 'presentation-page-layout-name'), None): cnv_StyleNameRef,
    ((PRESENTATIONNS, 'preset-class'), None): cnv_string,
    ((PRESENTATIONNS, 'preset-id'), None): cnv_string,
    ((PRESENTATIONNS, 'preset-sub-type'), None): cnv_string,
    ((PRESENTATIONNS, 'show'), None): cnv_string,
    ((PRESENTATIONNS, 'show-end-of-presentation-slide'), None): cnv_boolean,
    ((PRESENTATIONNS, 'show-logo'), None): cnv_boolean,
    ((PRESENTATIONNS, 'source'), None): cnv_string,
    ((PRESENTATIONNS, 'speed'), None): cnv_string,
    ((PRESENTATIONNS, 'start-page'), None): cnv_string,
    ((PRESENTATIONNS, 'start-scale'), None): cnv_string,
    ((PRESENTATIONNS, 'start-with-navigator'), None): cnv_boolean,
    ((PRESENTATIONNS, 'stay-on-top'), None): cnv_boolean,
    ((PRESENTATIONNS, 'style-name'), None): cnv_StyleNameRef,
    ((PRESENTATIONNS, 'transition-on-click'), None): cnv_string,
    ((PRESENTATIONNS, 'transition-speed'), None): cnv_string,
    ((PRESENTATIONNS, 'transition-style'), None): cnv_string,
    ((PRESENTATIONNS, 'transition-type'), None): cnv_string,
    ((PRESENTATIONNS, 'use-date-time-name'), None): cnv_string,
    ((PRESENTATIONNS, 'use-footer-name'), None): cnv_string,
    ((PRESENTATIONNS, 'use-header-name'), None): cnv_string,
    ((PRESENTATIONNS, 'user-transformed'), None): cnv_boolean,
    ((PRESENTATIONNS, 'verb'), None): cnv_nonNegativeInteger,
    ((PRESENTATIONNS, 'visibility'), None): cnv_string,
    ((SCRIPTNS, 'event-name'), None): cnv_formula,
    ((SCRIPTNS, 'language'), None): cnv_formula,
    ((SCRIPTNS, 'macro-name'), None): cnv_string,
    ((SMILNS, 'accelerate'), None): cnv_double,
    ((SMILNS, 'accumulate'), None): cnv_string,
    ((SMILNS, 'additive'), None): cnv_string,
    ((SMILNS, 'attributeName'), None): cnv_string,
    ((SMILNS, 'autoReverse'), None): cnv_boolean,
    ((SMILNS, 'begin'), None): cnv_string,
    ((SMILNS, 'by'), None): cnv_string,
    ((SMILNS, 'calcMode'), None): cnv_string,
    ((SMILNS, 'decelerate'), None): cnv_double,
    ((SMILNS, 'direction'), None): cnv_string,
    ((SMILNS, 'dur'), None): cnv_string,
    ((SMILNS, 'end'), None): cnv_string,
    ((SMILNS, 'endsync'), None): cnv_string,
    ((SMILNS, 'fadeColor'), None): cnv_string,
    ((SMILNS, 'fill'), None): cnv_string,
    ((SMILNS, 'fillDefault'), None): cnv_string,
    ((SMILNS, 'from'), None): cnv_string,
    ((SMILNS, 'keySplines'), None): cnv_string,
    ((SMILNS, 'keyTimes'), None): cnv_string,
    ((SMILNS, 'mode'), None): cnv_string,
    ((SMILNS, 'repeatCount'), None): cnv_nonNegativeInteger,
    ((SMILNS, 'repeatDur'), None): cnv_string,
    ((SMILNS, 'restart'), None): cnv_string,
    ((SMILNS, 'restartDefault'), None): cnv_string,
    ((SMILNS, 'subtype'), None): cnv_string,
    ((SMILNS, 'targetElement'), None): cnv_IDREF,
    ((SMILNS, 'to'), None): cnv_string,
    ((SMILNS, 'type'), None): cnv_string,
    ((SMILNS, 'values'), None): cnv_string,
    ((STYLENS, 'adjustment'), None): cnv_string,
    ((STYLENS, 'apply-style-name'), None): cnv_StyleNameRef,
    ((STYLENS, 'auto-text-indent'), None): cnv_boolean,
    ((STYLENS, 'auto-update'), None): cnv_boolean,
    ((STYLENS, 'background-transparency'), None): cnv_string,
    ((STYLENS, 'base-cell-address'), None): cnv_string,
    ((STYLENS, 'border-line-width-bottom'), None): cnv_string,
    ((STYLENS, 'border-line-width'), None): cnv_string,
    ((STYLENS, 'border-line-width-left'), None): cnv_string,
    ((STYLENS, 'border-line-width-right'), None): cnv_string,
    ((STYLENS, 'border-line-width-top'), None): cnv_string,
    ((STYLENS, 'cell-protect'), None): cnv_string,
    ((STYLENS, 'char'), None): cnv_string,
    ((STYLENS, 'class'), None): cnv_string,
    ((STYLENS, 'color'), None): cnv_string,
    ((STYLENS, 'column-width'), None): cnv_string,
    ((STYLENS, 'condition'), None): cnv_string,
    ((STYLENS, 'country-asian'), None): cnv_string,
    ((STYLENS, 'country-complex'), None): cnv_string,
    ((STYLENS, 'data-style-name'), None): cnv_StyleNameRef,
    ((STYLENS, 'decimal-places'), None): cnv_string,
    ((STYLENS, 'default-outline-level'), None): cnv_positiveInteger,
    ((STYLENS, 'diagonal-bl-tr'), None): cnv_string,
    ((STYLENS, 'diagonal-bl-tr-widths'), None): cnv_string,
    ((STYLENS, 'diagonal-tl-br'), None): cnv_string,
    ((STYLENS, 'diagonal-tl-br-widths'), None): cnv_string,
    ((STYLENS, 'direction'), None): cnv_string,
    ((STYLENS, 'display'), None): cnv_boolean,
    ((STYLENS, 'display-name'), None): cnv_string,
    ((STYLENS, 'distance-after-sep'), None): cnv_length,
    ((STYLENS, 'distance-before-sep'), None): cnv_length,
    ((STYLENS, 'distance'), None): cnv_length,
    ((STYLENS, 'dynamic-spacing'), None): cnv_boolean,
    ((STYLENS, 'editable'), None): cnv_boolean,
    ((STYLENS, 'family'), None): cnv_family,
    ((STYLENS, 'filter-name'), None): cnv_string,
    ((STYLENS, 'first-page-number'), None): cnv_string,
    ((STYLENS, 'flow-with-text'), None): cnv_boolean,
    ((STYLENS, 'font-adornments'), None): cnv_string,
    ((STYLENS, 'font-charset'), None): cnv_string,
    ((STYLENS, 'font-charset-asian'), None): cnv_string,
    ((STYLENS, 'font-charset-complex'), None): cnv_string,
    ((STYLENS, 'font-family-asian'), None): cnv_string,
    ((STYLENS, 'font-family-complex'), None): cnv_string,
    ((STYLENS, 'font-family-generic-asian'), None): cnv_string,
    ((STYLENS, 'font-family-generic'), None): cnv_string,
    ((STYLENS, 'font-family-generic-complex'), None): cnv_string,
    ((STYLENS, 'font-independent-line-spacing'), None): cnv_boolean,
    ((STYLENS, 'font-name-asian'), None): cnv_string,
    ((STYLENS, 'font-name'), None): cnv_string,
    ((STYLENS, 'font-name-complex'), None): cnv_string,
    ((STYLENS, 'font-pitch-asian'), None): cnv_string,
    ((STYLENS, 'font-pitch'), None): cnv_string,
    ((STYLENS, 'font-pitch-complex'), None): cnv_string,
    ((STYLENS, 'font-relief'), None): cnv_string,
    ((STYLENS, 'font-size-asian'), None): cnv_string,
    ((STYLENS, 'font-size-complex'), None): cnv_string,
    ((STYLENS, 'font-size-rel-asian'), None): cnv_length,
    ((STYLENS, 'font-size-rel'), None): cnv_length,
    ((STYLENS, 'font-size-rel-complex'), None): cnv_length,
    ((STYLENS, 'font-style-asian'), None): cnv_string,
    ((STYLENS, 'font-style-complex'), None): cnv_string,
    ((STYLENS, 'font-style-name-asian'), None): cnv_string,
    ((STYLENS, 'font-style-name'), None): cnv_string,
    ((STYLENS, 'font-style-name-complex'), None): cnv_string,
    ((STYLENS, 'font-weight-asian'), None): cnv_string,
    ((STYLENS, 'font-weight-complex'), None): cnv_string,
    ((STYLENS, 'footnote-max-height'), None): cnv_length,
    ((STYLENS, 'glyph-orientation-vertical'), None): cnv_string,
    ((STYLENS, 'height'), None): cnv_string,
    ((STYLENS, 'horizontal-pos'), None): cnv_string,
    ((STYLENS, 'horizontal-rel'), None): cnv_string,
    ((STYLENS, 'justify-single-word'), None): cnv_boolean,
    ((STYLENS, 'language-asian'), None): cnv_string,
    ((STYLENS, 'language-complex'), None): cnv_string,
    ((STYLENS, 'layout-grid-base-height'), None): cnv_length,
    ((STYLENS, 'layout-grid-color'), None): cnv_string,
    ((STYLENS, 'layout-grid-display'), None): cnv_boolean,
    ((STYLENS, 'layout-grid-lines'), None): cnv_string,
    ((STYLENS, 'layout-grid-mode'), None): cnv_string,
    ((STYLENS, 'layout-grid-print'), None): cnv_boolean,
    ((STYLENS, 'layout-grid-ruby-below'), None): cnv_boolean,
    ((STYLENS, 'layout-grid-ruby-height'), None): cnv_length,
    ((STYLENS, 'leader-char'), None): cnv_string,
    ((STYLENS, 'leader-color'), None): cnv_string,
    ((STYLENS, 'leader-style'), None): cnv_string,
    ((STYLENS, 'leader-text'), None): cnv_string,
    ((STYLENS, 'leader-text-style'), None): cnv_StyleNameRef,
    ((STYLENS, 'leader-type'), None): cnv_string,
    ((STYLENS, 'leader-width'), None): cnv_string,
    ((STYLENS, 'legend-expansion-aspect-ratio'), None): cnv_double,
    ((STYLENS, 'legend-expansion'), None): cnv_string,
    ((STYLENS, 'length'), None): cnv_positiveInteger,
    ((STYLENS, 'letter-kerning'), None): cnv_boolean,
    ((STYLENS, 'line-break'), None): cnv_string,
    ((STYLENS, 'line-height-at-least'), None): cnv_string,
    ((STYLENS, 'line-spacing'), None): cnv_length,
    ((STYLENS, 'line-style'), None): cnv_string,
    ((STYLENS, 'lines'), None): cnv_positiveInteger,
    ((STYLENS, 'list-style-name'), None): cnv_StyleNameRef,
    ((STYLENS, 'master-page-name'), None): cnv_StyleNameRef,
    ((STYLENS, 'may-break-between-rows'), None): cnv_boolean,
    ((STYLENS, 'min-row-height'), None): cnv_string,
    ((STYLENS, 'mirror'), None): cnv_string,
    ((STYLENS, 'name'), None): cnv_NCName,
    ((STYLENS, 'name'), (STYLENS, 'font-face')): cnv_string,
    ((STYLENS, 'next-style-name'), None): cnv_StyleNameRef,
    ((STYLENS, 'num-format'), None): cnv_string,
    ((STYLENS, 'num-letter-sync'), None): cnv_boolean,
    ((STYLENS, 'num-prefix'), None): cnv_string,
    ((STYLENS, 'num-suffix'), None): cnv_string,
    ((STYLENS, 'number-wrapped-paragraphs'), None): cnv_string,
    ((STYLENS, 'overflow-behavior'), None): cnv_string,
    ((STYLENS, 'page-layout-name'), None): cnv_StyleNameRef,
    ((STYLENS, 'page-number'), None): cnv_string,
    ((STYLENS, 'page-usage'), None): cnv_string,
    ((STYLENS, 'paper-tray-name'), None): cnv_string,
    ((STYLENS, 'parent-style-name'), None): cnv_StyleNameRef,
    ((STYLENS, 'position'), (STYLENS, 'tab-stop')): cnv_length,
    ((STYLENS, 'position'), None): cnv_string,
    ((STYLENS, 'print'), None): cnv_string,
    ((STYLENS, 'print-content'), None): cnv_boolean,
    ((STYLENS, 'print-orientation'), None): cnv_string,
    ((STYLENS, 'print-page-order'), None): cnv_string,
    ((STYLENS, 'protect'), None): cnv_boolean,
    ((STYLENS, 'punctuation-wrap'), None): cnv_string,
    ((STYLENS, 'register-true'), None): cnv_boolean,
    ((STYLENS, 'register-truth-ref-style-name'), None): cnv_string,
    ((STYLENS, 'rel-column-width'), None): cnv_string,
    ((STYLENS, 'rel-height'), None): cnv_string,
    ((STYLENS, 'rel-width'), None): cnv_string,
    ((STYLENS, 'repeat'), None): cnv_string,
    ((STYLENS, 'repeat-content'), None): cnv_boolean,
    ((STYLENS, 'rotation-align'), None): cnv_string,
    ((STYLENS, 'rotation-angle'), None): cnv_string,
    ((STYLENS, 'row-height'), None): cnv_string,
    ((STYLENS, 'ruby-align'), None): cnv_string,
    ((STYLENS, 'ruby-position'), None): cnv_string,
    ((STYLENS, 'run-through'), None): cnv_string,
    ((STYLENS, 'scale-to'), None): cnv_string,
    ((STYLENS, 'scale-to-pages'), None): cnv_string,
    ((STYLENS, 'script-type'), None): cnv_string,
    ((STYLENS, 'shadow'), None): cnv_string,
    ((STYLENS, 'shrink-to-fit'), None): cnv_boolean,
    ((STYLENS, 'snap-to-layout-grid'), None): cnv_boolean,
    ((STYLENS, 'style'), None): cnv_string,
    ((STYLENS, 'style-name'), None): cnv_StyleNameRef,
    ((STYLENS, 'tab-stop-distance'), None): cnv_string,
    ((STYLENS, 'table-centering'), None): cnv_string,
    ((STYLENS, 'text-align-source'), None): cnv_string,
    ((STYLENS, 'text-autospace'), None): cnv_string,
    ((STYLENS, 'text-blinking'), None): cnv_boolean,
    ((STYLENS, 'text-combine'), None): cnv_string,
    ((STYLENS, 'text-combine-end-char'), None): cnv_string,
    ((STYLENS, 'text-combine-start-char'), None): cnv_string,
    ((STYLENS, 'text-emphasize'), None): cnv_string,
    ((STYLENS, 'text-line-through-color'), None): cnv_string,
    ((STYLENS, 'text-line-through-mode'), None): cnv_string,
    ((STYLENS, 'text-line-through-style'), None): cnv_string,
    ((STYLENS, 'text-line-through-text'), None): cnv_string,
    ((STYLENS, 'text-line-through-text-style'), None): cnv_string,
    ((STYLENS, 'text-line-through-type'), None): cnv_string,
    ((STYLENS, 'text-line-through-width'), None): cnv_string,
    ((STYLENS, 'text-outline'), None): cnv_boolean,
    ((STYLENS, 'text-position'), None): cnv_string,
    ((STYLENS, 'text-rotation-angle'), None): cnv_string,
    ((STYLENS, 'text-rotation-scale'), None): cnv_string,
    ((STYLENS, 'text-scale'), None): cnv_string,
    ((STYLENS, 'text-underline-color'), None): cnv_string,
    ((STYLENS, 'text-underline-mode'), None): cnv_string,
    ((STYLENS, 'text-underline-style'), None): cnv_string,
    ((STYLENS, 'text-underline-type'), None): cnv_string,
    ((STYLENS, 'text-underline-width'), None): cnv_string,
    ((STYLENS, 'type'), None): cnv_string,
    ((STYLENS, 'use-optimal-column-width'), None): cnv_boolean,
    ((STYLENS, 'use-optimal-row-height'), None): cnv_boolean,
    ((STYLENS, 'use-window-font-color'), None): cnv_boolean,
    ((STYLENS, 'vertical-align'), None): cnv_string,
    ((STYLENS, 'vertical-pos'), None): cnv_string,
    ((STYLENS, 'vertical-rel'), None): cnv_string,
    ((STYLENS, 'volatile'), None): cnv_boolean,
    ((STYLENS, 'width'), None): cnv_string,
    ((STYLENS, 'wrap'), None): cnv_string,
    ((STYLENS, 'wrap-contour'), None): cnv_boolean,
    ((STYLENS, 'wrap-contour-mode'), None): cnv_string,
    ((STYLENS, 'wrap-dynamic-threshold'), None): cnv_length,
    ((STYLENS, 'writing-mode-automatic'), None): cnv_boolean,
    ((STYLENS, 'writing-mode'), None): cnv_string,
    ((SVGNS, 'accent-height'), None): cnv_integer,
    ((SVGNS, 'alphabetic'), None): cnv_integer,
    ((SVGNS, 'ascent'), None): cnv_integer,
    ((SVGNS, 'bbox'), None): cnv_string,
    ((SVGNS, 'cap-height'), None): cnv_integer,
    ((SVGNS, 'cx'), None): cnv_string,
    ((SVGNS, 'cy'), None): cnv_string,
    ((SVGNS, 'd'), None): cnv_string,
    ((SVGNS, 'descent'), None): cnv_integer,
    ((SVGNS, 'fill-rule'), None): cnv_string,
    ((SVGNS, 'font-family'), None): cnv_string,
    ((SVGNS, 'font-size'), None): cnv_string,
    ((SVGNS, 'font-stretch'), None): cnv_string,
    ((SVGNS, 'font-style'), None): cnv_string,
    ((SVGNS, 'font-variant'), None): cnv_string,
    ((SVGNS, 'font-weight'), None): cnv_string,
    ((SVGNS, 'fx'), None): cnv_string,
    ((SVGNS, 'fy'), None): cnv_string,
    ((SVGNS, 'gradientTransform'), None): cnv_string,
    ((SVGNS, 'gradientUnits'), None): cnv_string,
    ((SVGNS, 'hanging'), None): cnv_integer,
    ((SVGNS, 'height'), None): cnv_length,
    ((SVGNS, 'ideographic'), None): cnv_integer,
    ((SVGNS, 'mathematical'), None): cnv_integer,
    ((SVGNS, 'name'), None): cnv_string,
    ((SVGNS, 'offset'), None): cnv_string,
    ((SVGNS, 'origin'), None): cnv_string,
    ((SVGNS, 'overline-position'), None): cnv_integer,
    ((SVGNS, 'overline-thickness'), None): cnv_integer,
    ((SVGNS, 'panose-1'), None): cnv_string,
    ((SVGNS, 'path'), None): cnv_string,
    ((SVGNS, 'r'), None): cnv_length,
    ((SVGNS, 'rx'), None): cnv_length,
    ((SVGNS, 'ry'), None): cnv_length,
    ((SVGNS, 'slope'), None): cnv_integer,
    ((SVGNS, 'spreadMethod'), None): cnv_string,
    ((SVGNS, 'stemh'), None): cnv_integer,
    ((SVGNS, 'stemv'), None): cnv_integer,
    ((SVGNS, 'stop-color'), None): cnv_string,
    ((SVGNS, 'stop-opacity'), None): cnv_double,
    ((SVGNS, 'strikethrough-position'), None): cnv_integer,
    ((SVGNS, 'strikethrough-thickness'), None): cnv_integer,
    ((SVGNS, 'string'), None): cnv_string,
    ((SVGNS, 'stroke-color'), None): cnv_string,
    ((SVGNS, 'stroke-opacity'), None): cnv_string,
    ((SVGNS, 'stroke-width'), None): cnv_length,
    ((SVGNS, 'type'), None): cnv_string,
    ((SVGNS, 'underline-position'), None): cnv_integer,
    ((SVGNS, 'underline-thickness'), None): cnv_integer,
    ((SVGNS, 'unicode-range'), None): cnv_string,
    ((SVGNS, 'units-per-em'), None): cnv_integer,
    ((SVGNS, 'v-alphabetic'), None): cnv_integer,
    ((SVGNS, 'v-hanging'), None): cnv_integer,
    ((SVGNS, 'v-ideographic'), None): cnv_integer,
    ((SVGNS, 'v-mathematical'), None): cnv_integer,
    ((SVGNS, 'viewBox'), None): cnv_viewbox,
    ((SVGNS, 'width'), None): cnv_length,
    ((SVGNS, 'widths'), None): cnv_string,
    ((SVGNS, 'x'), None): cnv_length,
    ((SVGNS, 'x-height'), None): cnv_integer,
    ((SVGNS, 'x1'), None): cnv_lengthorpercent,
    ((SVGNS, 'x2'), None): cnv_lengthorpercent,
    ((SVGNS, 'y'), None): cnv_length,
    ((SVGNS, 'y1'), None): cnv_lengthorpercent,
    ((SVGNS, 'y2'), None): cnv_lengthorpercent,
    ((TABLENS, 'acceptance-state'), None): cnv_string,
    ((TABLENS, 'add-empty-lines'), None): cnv_boolean,
    ((TABLENS, 'algorithm'), None): cnv_formula,
    ((TABLENS, 'align'), None): cnv_string,
    ((TABLENS, 'allow-empty-cell'), None): cnv_boolean,
    ((TABLENS, 'application-data'), None): cnv_string,
    ((TABLENS, 'automatic-find-labels'), None): cnv_boolean,
    ((TABLENS, 'base-cell-address'), None): cnv_string,
    ((TABLENS, 'bind-styles-to-content'), None): cnv_boolean,
    ((TABLENS, 'border-color'), None): cnv_string,
    ((TABLENS, 'border-model'), None): cnv_string,
    ((TABLENS, 'buttons'), None): cnv_string,
    # ((TABLENS,'case-sensitive'), None): cnv_boolean,
    ((TABLENS, 'case-sensitive'), None): cnv_string,
    ((TABLENS, 'cell-address'), None): cnv_string,
    ((TABLENS, 'cell-range-address'), None): cnv_string,
    ((TABLENS, 'cell-range'), None): cnv_string,
    ((TABLENS, 'column'), None): cnv_integer,
    ((TABLENS, 'comment'), None): cnv_string,
    ((TABLENS, 'condition'), None): cnv_formula,
    ((TABLENS, 'condition-source'), None): cnv_string,
    ((TABLENS, 'condition-source-range-address'), None): cnv_string,
    ((TABLENS, 'contains-error'), None): cnv_boolean,
    ((TABLENS, 'contains-header'), None): cnv_boolean,
    ((TABLENS, 'content-validation-name'), None): cnv_string,
    ((TABLENS, 'copy-back'), None): cnv_boolean,
    ((TABLENS, 'copy-formulas'), None): cnv_boolean,
    ((TABLENS, 'copy-styles'), None): cnv_boolean,
    ((TABLENS, 'count'), None): cnv_positiveInteger,
    ((TABLENS, 'country'), None): cnv_token,
    ((TABLENS, 'data-cell-range-address'), None): cnv_string,
    ((TABLENS, 'data-field'), None): cnv_string,
    ((TABLENS, 'data-type'), None): cnv_string,
    ((TABLENS, 'database-name'), None): cnv_string,
    ((TABLENS, 'database-table-name'), None): cnv_string,
    ((TABLENS, 'date-end'), None): cnv_string,
    ((TABLENS, 'date-start'), None): cnv_string,
    ((TABLENS, 'date-value'), None): cnv_date,
    ((TABLENS, 'default-cell-style-name'), None): cnv_StyleNameRef,
    ((TABLENS, 'direction'), None): cnv_string,
    ((TABLENS, 'display-border'), None): cnv_boolean,
    ((TABLENS, 'display'), None): cnv_boolean,
    ((TABLENS, 'display-duplicates'), None): cnv_boolean,
    ((TABLENS, 'display-filter-buttons'), None): cnv_boolean,
    ((TABLENS, 'display-list'), None): cnv_string,
    ((TABLENS, 'display-member-mode'), None): cnv_string,
    ((TABLENS, 'drill-down-on-double-click'), None): cnv_boolean,
    ((TABLENS, 'enabled'), None): cnv_boolean,
    ((TABLENS, 'end-cell-address'), None): cnv_string,
    ((TABLENS, 'end'), None): cnv_string,
    ((TABLENS, 'end-column'), None): cnv_integer,
    ((TABLENS, 'end-position'), None): cnv_integer,
    ((TABLENS, 'end-row'), None): cnv_integer,
    ((TABLENS, 'end-table'), None): cnv_integer,
    ((TABLENS, 'end-x'), None): cnv_length,
    ((TABLENS, 'end-y'), None): cnv_length,
    ((TABLENS, 'execute'), None): cnv_boolean,
    ((TABLENS, 'expression'), None): cnv_formula,
    ((TABLENS, 'field-name'), None): cnv_string,
    # ((TABLENS,'field-number'), None): cnv_nonNegativeInteger,
    ((TABLENS, 'field-number'), None): cnv_string,
    ((TABLENS, 'filter-name'), None): cnv_string,
    ((TABLENS, 'filter-options'), None): cnv_string,
    ((TABLENS, 'formula'), None): cnv_formula,
    ((TABLENS, 'function'), None): cnv_string,
    ((TABLENS, 'grand-total'), None): cnv_string,
    ((TABLENS, 'group-by-field-number'), None): cnv_nonNegativeInteger,
    ((TABLENS, 'grouped-by'), None): cnv_string,
    ((TABLENS, 'has-persistent-data'), None): cnv_boolean,
    ((TABLENS, 'id'), None): cnv_string,
    ((TABLENS, 'identify-categories'), None): cnv_boolean,
    ((TABLENS, 'ignore-empty-rows'), None): cnv_boolean,
    ((TABLENS, 'index'), None): cnv_nonNegativeInteger,
    ((TABLENS, 'is-active'), None): cnv_boolean,
    ((TABLENS, 'is-data-layout-field'), None): cnv_string,
    ((TABLENS, 'is-selection'), None): cnv_boolean,
    ((TABLENS, 'is-sub-table'), None): cnv_boolean,
    ((TABLENS, 'label-cell-range-address'), None): cnv_string,
    ((TABLENS, 'language'), None): cnv_token,
    ((TABLENS, 'last-column-spanned'), None): cnv_positiveInteger,
    ((TABLENS, 'last-row-spanned'), None): cnv_positiveInteger,
    ((TABLENS, 'layout-mode'), None): cnv_string,
    ((TABLENS, 'link-to-source-data'), None): cnv_boolean,
    ((TABLENS, 'marked-invalid'), None): cnv_boolean,
    ((TABLENS, 'matrix-covered'), None): cnv_boolean,
    ((TABLENS, 'maximum-difference'), None): cnv_double,
    ((TABLENS, 'member-count'), None): cnv_nonNegativeInteger,
    ((TABLENS, 'member-name'), None): cnv_string,
    ((TABLENS, 'member-type'), None): cnv_string,
    ((TABLENS, 'message-type'), None): cnv_string,
    ((TABLENS, 'mode'), None): cnv_string,
    ((TABLENS, 'multi-deletion-spanned'), None): cnv_integer,
    ((TABLENS, 'name'), None): cnv_string,
    ((TABLENS, 'null-year'), None): cnv_positiveInteger,
    ((TABLENS, 'number-columns-repeated'), None): cnv_positiveInteger,
    ((TABLENS, 'number-columns-spanned'), None): cnv_positiveInteger,
    ((TABLENS, 'number-matrix-columns-spanned'), None): cnv_positiveInteger,
    ((TABLENS, 'number-matrix-rows-spanned'), None): cnv_positiveInteger,
    ((TABLENS, 'number-rows-repeated'), None): cnv_positiveInteger,
    ((TABLENS, 'number-rows-spanned'), None): cnv_positiveInteger,
    ((TABLENS, 'object-name'), None): cnv_string,
    ((TABLENS, 'on-update-keep-size'), None): cnv_boolean,
    ((TABLENS, 'on-update-keep-styles'), None): cnv_boolean,
    ((TABLENS, 'operator'), None): cnv_string,
    ((TABLENS, 'order'), None): cnv_string,
    ((TABLENS, 'orientation'), None): cnv_string,
    ((TABLENS, 'page-breaks-on-group-change'), None): cnv_boolean,
    ((TABLENS, 'parse-sql-statement'), None): cnv_boolean,
    ((TABLENS, 'password'), None): cnv_string,
    ((TABLENS, 'position'), None): cnv_integer,
    ((TABLENS, 'precision-as-shown'), None): cnv_boolean,
    ((TABLENS, 'print'), None): cnv_boolean,
    ((TABLENS, 'print-ranges'), None): cnv_string,
    ((TABLENS, 'protect'), None): cnv_boolean,
    ((TABLENS, 'protected'), None): cnv_boolean,
    ((TABLENS, 'protection-key'), None): cnv_string,
    ((TABLENS, 'query-name'), None): cnv_string,
    ((TABLENS, 'range-usable-as'), None): cnv_string,
    # ((TABLENS,'refresh-delay'), None): cnv_boolean,
    ((TABLENS, 'refresh-delay'), None): cnv_duration,
    ((TABLENS, 'rejecting-change-id'), None): cnv_string,
    ((TABLENS, 'row'), None): cnv_integer,
    ((TABLENS, 'scenario-ranges'), None): cnv_string,
    ((TABLENS, 'search-criteria-must-apply-to-whole-cell'), None): cnv_boolean,
    ((TABLENS, 'selected-page'), None): cnv_string,
    ((TABLENS, 'show-details'), None): cnv_boolean,
    # ((TABLENS,'show-empty'), None): cnv_boolean,
    ((TABLENS, 'show-empty'), None): cnv_string,
    ((TABLENS, 'show-filter-button'), None): cnv_boolean,
    ((TABLENS, 'sort-mode'), None): cnv_string,
    ((TABLENS, 'source-cell-range-addresses'), None): cnv_string,
    ((TABLENS, 'source-field-name'), None): cnv_string,
    ((TABLENS, 'source-name'), None): cnv_string,
    ((TABLENS, 'sql-statement'), None): cnv_string,
    ((TABLENS, 'start'), None): cnv_string,
    ((TABLENS, 'start-column'), None): cnv_integer,
    ((TABLENS, 'start-position'), None): cnv_integer,
    ((TABLENS, 'start-row'), None): cnv_integer,
    ((TABLENS, 'start-table'), None): cnv_integer,
    ((TABLENS, 'status'), None): cnv_string,
    ((TABLENS, 'step'), None): cnv_double,
    ((TABLENS, 'steps'), None): cnv_positiveInteger,
    ((TABLENS, 'structure-protected'), None): cnv_boolean,
    ((TABLENS, 'style-name'), None): cnv_StyleNameRef,
    ((TABLENS, 'table-background'), None): cnv_boolean,
    ((TABLENS, 'table'), None): cnv_integer,
    ((TABLENS, 'table-name'), None): cnv_string,
    ((TABLENS, 'target-cell-address'), None): cnv_string,
    ((TABLENS, 'target-range-address'), None): cnv_string,
    ((TABLENS, 'title'), None): cnv_string,
    ((TABLENS, 'track-changes'), None): cnv_boolean,
    ((TABLENS, 'type'), None): cnv_string,
    ((TABLENS, 'use-labels'), None): cnv_string,
    ((TABLENS, 'use-regular-expressions'), None): cnv_boolean,
    ((TABLENS, 'used-hierarchy'), None): cnv_integer,
    ((TABLENS, 'user-name'), None): cnv_string,
    ((TABLENS, 'value'), None): cnv_string,
    ((TABLENS, 'value-type'), None): cnv_string,
    ((TABLENS, 'visibility'), None): cnv_string,
    ((TEXTNS, 'active'), None): cnv_boolean,
    ((TEXTNS, 'address'), None): cnv_string,
    ((TEXTNS, 'alphabetical-separators'), None): cnv_boolean,
    ((TEXTNS, 'anchor-page-number'), None): cnv_positiveInteger,
    ((TEXTNS, 'anchor-type'), None): cnv_string,
    ((TEXTNS, 'animation'), None): cnv_string,
    ((TEXTNS, 'animation-delay'), None): cnv_string,
    ((TEXTNS, 'animation-direction'), None): cnv_string,
    ((TEXTNS, 'animation-repeat'), None): cnv_string,
    ((TEXTNS, 'animation-start-inside'), None): cnv_boolean,
    ((TEXTNS, 'animation-steps'), None): cnv_length,
    ((TEXTNS, 'animation-stop-inside'), None): cnv_boolean,
    ((TEXTNS, 'annote'), None): cnv_string,
    ((TEXTNS, 'author'), None): cnv_string,
    ((TEXTNS, 'bibliography-data-field'), None): cnv_string,
    ((TEXTNS, 'bibliography-type'), None): cnv_string,
    ((TEXTNS, 'booktitle'), None): cnv_string,
    ((TEXTNS, 'bullet-char'), None): cnv_string,
    ((TEXTNS, 'bullet-relative-size'), None): cnv_string,
    ((TEXTNS, 'c'), None): cnv_nonNegativeInteger,
    ((TEXTNS, 'capitalize-entries'), None): cnv_boolean,
    ((TEXTNS, 'caption-sequence-format'), None): cnv_string,
    ((TEXTNS, 'caption-sequence-name'), None): cnv_string,
    ((TEXTNS, 'change-id'), None): cnv_IDREF,
    ((TEXTNS, 'chapter'), None): cnv_string,
    ((TEXTNS, 'citation-body-style-name'), None): cnv_StyleNameRef,
    ((TEXTNS, 'citation-style-name'), None): cnv_StyleNameRef,
    ((TEXTNS, 'class-names'), None): cnv_NCNames,
    ((TEXTNS, 'column-name'), None): cnv_string,
    ((TEXTNS, 'combine-entries'), None): cnv_boolean,
    ((TEXTNS, 'combine-entries-with-dash'), None): cnv_boolean,
    ((TEXTNS, 'combine-entries-with-pp'), None): cnv_boolean,
    ((TEXTNS, 'comma-separated'), None): cnv_boolean,
    ((TEXTNS, 'cond-style-name'), None): cnv_StyleNameRef,
    ((TEXTNS, 'condition'), None): cnv_formula,
    ((TEXTNS, 'connection-name'), None): cnv_string,
    ((TEXTNS, 'consecutive-numbering'), None): cnv_boolean,
    ((TEXTNS, 'continue-numbering'), None): cnv_boolean,
    ((TEXTNS, 'copy-outline-levels'), None): cnv_boolean,
    ((TEXTNS, 'count-empty-lines'), None): cnv_boolean,
    ((TEXTNS, 'count-in-text-boxes'), None): cnv_boolean,
    ((TEXTNS, 'current-value'), None): cnv_boolean,
    ((TEXTNS, 'custom1'), None): cnv_string,
    ((TEXTNS, 'custom2'), None): cnv_string,
    ((TEXTNS, 'custom3'), None): cnv_string,
    ((TEXTNS, 'custom4'), None): cnv_string,
    ((TEXTNS, 'custom5'), None): cnv_string,
    ((TEXTNS, 'database-name'), None): cnv_string,
    ((TEXTNS, 'date-adjust'), None): cnv_duration,
    ((TEXTNS, 'date-value'), None): cnv_date,
    # ((TEXTNS,'date-value'), None): cnv_dateTime,
    ((TEXTNS, 'default-style-name'), None): cnv_StyleNameRef,
    ((TEXTNS, 'description'), None): cnv_string,
    ((TEXTNS, 'display'), None): cnv_string,
    ((TEXTNS, 'display-levels'), None): cnv_positiveInteger,
    ((TEXTNS, 'display-outline-level'), None): cnv_nonNegativeInteger,
    ((TEXTNS, 'dont-balance-text-columns'), None): cnv_boolean,
    ((TEXTNS, 'duration'), None): cnv_duration,
    ((TEXTNS, 'edition'), None): cnv_string,
    ((TEXTNS, 'editor'), None): cnv_string,
    ((TEXTNS, 'filter-name'), None): cnv_string,
    ((TEXTNS, 'first-row-end-column'), None): cnv_string,
    ((TEXTNS, 'first-row-start-column'), None): cnv_string,
    ((TEXTNS, 'fixed'), None): cnv_boolean,
    ((TEXTNS, 'footnotes-position'), None): cnv_string,
    ((TEXTNS, 'formula'), None): cnv_formula,
    ((TEXTNS, 'global'), None): cnv_boolean,
    ((TEXTNS, 'howpublished'), None): cnv_string,
    ((TEXTNS, 'id'), None): cnv_ID,
    # ((TEXTNS,'id'), None): cnv_string,
    ((TEXTNS, 'identifier'), None): cnv_string,
    ((TEXTNS, 'ignore-case'), None): cnv_boolean,
    ((TEXTNS, 'increment'), None): cnv_nonNegativeInteger,
    ((TEXTNS, 'index-name'), None): cnv_string,
    ((TEXTNS, 'index-scope'), None): cnv_string,
    ((TEXTNS, 'institution'), None): cnv_string,
    ((TEXTNS, 'is-hidden'), None): cnv_boolean,
    ((TEXTNS, 'is-list-header'), None): cnv_boolean,
    ((TEXTNS, 'isbn'), None): cnv_string,
    ((TEXTNS, 'issn'), None): cnv_string,
    ((TEXTNS, 'journal'), None): cnv_string,
    ((TEXTNS, 'key'), None): cnv_string,
    ((TEXTNS, 'key1'), None): cnv_string,
    ((TEXTNS, 'key1-phonetic'), None): cnv_string,
    ((TEXTNS, 'key2'), None): cnv_string,
    ((TEXTNS, 'key2-phonetic'), None): cnv_string,
    ((TEXTNS, 'kind'), None): cnv_string,
    ((TEXTNS, 'label'), None): cnv_string,
    ((TEXTNS, 'last-row-end-column'), None): cnv_string,
    ((TEXTNS, 'last-row-start-column'), None): cnv_string,
    ((TEXTNS, 'level'), None): cnv_positiveInteger,
    ((TEXTNS, 'line-break'), None): cnv_boolean,
    ((TEXTNS, 'line-number'), None): cnv_string,
    ((TEXTNS, 'main-entry'), None): cnv_boolean,
    ((TEXTNS, 'main-entry-style-name'), None): cnv_StyleNameRef,
    ((TEXTNS, 'master-page-name'), None): cnv_StyleNameRef,
    ((TEXTNS, 'min-label-distance'), None): cnv_string,
    ((TEXTNS, 'min-label-width'), None): cnv_string,
    ((TEXTNS, 'month'), None): cnv_string,
    ((TEXTNS, 'name'), None): cnv_string,
    ((TEXTNS, 'note-class'), None): cnv_textnoteclass,
    ((TEXTNS, 'note'), None): cnv_string,
    ((TEXTNS, 'number'), None): cnv_string,
    ((TEXTNS, 'number-lines'), None): cnv_boolean,
    ((TEXTNS, 'number-position'), None): cnv_string,
    ((TEXTNS, 'numbered-entries'), None): cnv_boolean,
    ((TEXTNS, 'offset'), None): cnv_string,
    ((TEXTNS, 'organizations'), None): cnv_string,
    ((TEXTNS, 'outline-level'), None): cnv_string,
    ((TEXTNS, 'page-adjust'), None): cnv_integer,
    ((TEXTNS, 'pages'), None): cnv_string,
    ((TEXTNS, 'paragraph-style-name'), None): cnv_StyleNameRef,
    ((TEXTNS, 'placeholder-type'), None): cnv_string,
    ((TEXTNS, 'prefix'), None): cnv_string,
    ((TEXTNS, 'protected'), None): cnv_boolean,
    ((TEXTNS, 'protection-key'), None): cnv_string,
    ((TEXTNS, 'publisher'), None): cnv_string,
    ((TEXTNS, 'ref-name'), None): cnv_string,
    ((TEXTNS, 'reference-format'), None): cnv_string,
    ((TEXTNS, 'relative-tab-stop-position'), None): cnv_boolean,
    ((TEXTNS, 'report-type'), None): cnv_string,
    ((TEXTNS, 'restart-numbering'), None): cnv_boolean,
    ((TEXTNS, 'restart-on-page'), None): cnv_boolean,
    ((TEXTNS, 'row-number'), None): cnv_nonNegativeInteger,
    ((TEXTNS, 'school'), None): cnv_string,
    ((TEXTNS, 'section-name'), None): cnv_string,
    ((TEXTNS, 'select-page'), None): cnv_string,
    ((TEXTNS, 'separation-character'), None): cnv_string,
    ((TEXTNS, 'series'), None): cnv_string,
    ((TEXTNS, 'sort-algorithm'), None): cnv_string,
    ((TEXTNS, 'sort-ascending'), None): cnv_boolean,
    ((TEXTNS, 'sort-by-position'), None): cnv_boolean,
    ((TEXTNS, 'space-before'), None): cnv_string,
    ((TEXTNS, 'start-numbering-at'), None): cnv_string,
    # ((TEXTNS,'start-value'), None): cnv_nonNegativeInteger,
    ((TEXTNS, 'start-value'), None): cnv_positiveInteger,
    ((TEXTNS, 'string-value'), None): cnv_string,
    ((TEXTNS, 'string-value-if-false'), None): cnv_string,
    ((TEXTNS, 'string-value-if-true'), None): cnv_string,
    ((TEXTNS, 'string-value-phonetic'), None): cnv_string,
    ((TEXTNS, 'style-name'), None): cnv_StyleNameRef,
    ((TEXTNS, 'suffix'), None): cnv_string,
    ((TEXTNS, 'tab-ref'), None): cnv_nonNegativeInteger,
    ((TEXTNS, 'table-name'), None): cnv_string,
    ((TEXTNS, 'table-type'), None): cnv_string,
    ((TEXTNS, 'time-adjust'), None): cnv_duration,
    # ((TEXTNS,'time-value'), None): cnv_dateTime,
    ((TEXTNS, 'time-value'), None): cnv_time,
    ((TEXTNS, 'title'), None): cnv_string,
    ((TEXTNS, 'track-changes'), None): cnv_boolean,
    ((TEXTNS, 'url'), None): cnv_string,
    ((TEXTNS, 'use-caption'), None): cnv_boolean,
    ((TEXTNS, 'use-chart-objects'), None): cnv_boolean,
    ((TEXTNS, 'use-draw-objects'), None): cnv_boolean,
    ((TEXTNS, 'use-floating-frames'), None): cnv_boolean,
    ((TEXTNS, 'use-graphics'), None): cnv_boolean,
    ((TEXTNS, 'use-index-marks'), None): cnv_boolean,
    ((TEXTNS, 'use-index-source-styles'), None): cnv_boolean,
    ((TEXTNS, 'use-keys-as-entries'), None): cnv_boolean,
    ((TEXTNS, 'use-math-objects'), None): cnv_boolean,
    ((TEXTNS, 'use-objects'), None): cnv_boolean,
    ((TEXTNS, 'use-other-objects'), None): cnv_boolean,
    ((TEXTNS, 'use-outline-level'), None): cnv_boolean,
    ((TEXTNS, 'use-soft-page-breaks'), None): cnv_boolean,
    ((TEXTNS, 'use-spreadsheet-objects'), None): cnv_boolean,
    ((TEXTNS, 'use-tables'), None): cnv_boolean,
    ((TEXTNS, 'value'), None): cnv_nonNegativeInteger,
    ((TEXTNS, 'visited-style-name'), None): cnv_StyleNameRef,
    ((TEXTNS, 'volume'), None): cnv_string,
    ((TEXTNS, 'year'), None): cnv_string,
    ((XFORMSNS, 'bind'), None): cnv_string,
    ((XLINKNS, 'actuate'), None): cnv_string,
    ((XLINKNS, 'href'), None): cnv_anyURI,
    ((XLINKNS, 'show'), None): cnv_xlinkshow,
    ((XLINKNS, 'title'), None): cnv_string,
    ((XLINKNS, 'type'), None): cnv_string,
}


class AttrConverters:
    def convert(self, attribute, value, element):
        """Based on the element, figures out how to check/convert the attribute value
        All values are converted to string
        """
        conversion = attrconverters.get((attribute, element.qname), None)
        if conversion is not None:
            return conversion(attribute, value, element)
        else:
            conversion = attrconverters.get((attribute, None), None)
            if conversion is not None:
                return conversion(attribute, value, element)
        return str(value)



#!/usr/bin/env python
# License: GPLv3 Copyright: 2013, Kovid Goyal <kovid at kovidgoyal.net>

import os
import re
import time
from collections import defaultdict
from contextlib import contextmanager, suppress
from functools import partial
from typing import Any

from calibre import prints
from calibre.constants import filesystem_encoding, ismacos, iswindows
from calibre.ebooks import BOOK_EXTENSIONS
from calibre.utils.filenames import make_long_path_useable
from calibre.utils.icu import lower as icu_lower
from calibre.utils.localization import _


def splitext(path):
    key, ext = os.path.splitext(path)
    return key, ext[1:].lower()


def formats_ok(formats):
    return len(formats) > 0


def path_ok(path):
    return not os.path.isdir(path) and os.access(path, os.R_OK)


def compile_glob(pat):
    import fnmatch

    return re.compile(fnmatch.translate(pat), flags=re.I)


def compile_rule(rule):
    mt = rule['match_type']
    if 'with' in mt:
        q = icu_lower(rule['query'])
        if 'startswith' in mt:

            def func(filename):
                return icu_lower(filename).startswith(q)

        else:

            def func(filename):
                return icu_lower(filename).endswith(q)

    elif 'glob' in mt:
        q = compile_glob(rule['query'])

        def func(filename):
            return q.match(filename) is not None

    else:
        q = re.compile(rule['query'])

        def func(filename):
            return q.match(filename) is not None

    ans = func
    if mt.startswith('not_'):

        def ans(filename):
            return not func(filename)

    return ans, rule['action'] == 'add'


def filter_filename(compiled_rules, filename):
    for q, action in compiled_rules:
        if q(filename):
            return action


_metadata_extensions = None


def metadata_extensions():
    # Set of all known book extensions + OPF (the OPF is used to read metadata,
    # but not actually added)
    global _metadata_extensions
    if _metadata_extensions is None:
        _metadata_extensions = frozenset(BOOK_EXTENSIONS) | {'opf'}
    return _metadata_extensions


if iswindows or ismacos:
    unicode_listdir = os.listdir
else:

    def unicode_listdir(root):
        root = root.encode(filesystem_encoding)
        for x in os.listdir(root):
            try:
                yield x.decode(filesystem_encoding)
            except UnicodeDecodeError:
                prints('Ignoring un-decodable file:', x)


def listdir(root, sort_by_mtime=False):
    items = (make_long_path_useable(os.path.join(root, x)) for x in unicode_listdir(root))
    if sort_by_mtime:

        def safe_mtime(x):
            try:
                return os.path.getmtime(x)
            except OSError:
                return time.time()

        items = sorted(items, key=safe_mtime)

    for path in items:
        if path_ok(path):
            yield path


def list_only_files_in_dir(root, sort_by_mtime=False):
    def files_iter():
        for x in os.scandir(root):
            with suppress(OSError):
                if x.is_file(follow_symlinks=True):
                    yield x

    items = files_iter()
    if sort_by_mtime:

        def safe_mtime(x: os.DirEntry):
            with suppress(OSError):
                return x.stat(follow_symlinks=True).st_mtime_ns
            return 0

        items = sorted(items, key=safe_mtime)
    yield from (make_long_path_useable(os.path.join(root, x.name)) for x in items)


def allow_path(path, ext, compiled_rules):
    ans = filter_filename(compiled_rules, os.path.basename(path))
    if ans is None:
        ans = ext in metadata_extensions()
    return ans


import_ctx: dict[str, Any] | None = None


@contextmanager
def run_import_plugins_before_metadata(tdir, group_id=0):
    global import_ctx
    import_ctx = {'tdir': tdir, 'group_id': group_id, 'format_map': {}}
    yield import_ctx
    import_ctx = None


def run_import_plugins(formats):
    from calibre.ebooks.metadata.worker import run_import_plugins

    assert import_ctx is not None
    import_ctx['group_id'] += 1
    ans = run_import_plugins(formats, import_ctx['group_id'], import_ctx['tdir'])
    fm = import_ctx['format_map']
    for old_path, new_path in zip(formats, ans):
        new_path = make_long_path_useable(new_path)
        fm[new_path] = old_path
    return ans


def find_books_in_directory(dirpath, single_book_per_directory, compiled_rules=(), listdir_impl=list_only_files_in_dir):
    dirpath = make_long_path_useable(os.path.abspath(dirpath))
    if single_book_per_directory:
        formats = {}
        for path in listdir_impl(dirpath):
            key, ext = splitext(path)
            if allow_path(path, ext, compiled_rules):
                formats[ext] = path
        if formats_ok(formats):
            yield list(formats.values())
    else:
        books = defaultdict(dict)
        for path in listdir_impl(dirpath, sort_by_mtime=True):
            key, ext = splitext(path)
            if allow_path(path, ext, compiled_rules):
                books[icu_lower(key) if isinstance(key, str) else key.lower()][ext] = path

        for formats in books.values():
            if formats_ok(formats):
                yield list(formats.values())


def create_format_map(formats):
    format_map = {}
    for path in formats:
        ext = os.path.splitext(path)[1][1:].upper()
        if ext == 'OPF':
            continue
        format_map[ext] = path
    return format_map


def import_book_directory_multiple(db, dirpath, callback=None, added_ids=None, compiled_rules=(), add_duplicates=False):
    from calibre.ebooks.metadata.meta import metadata_from_formats

    duplicates = []
    for formats in find_books_in_directory(dirpath, False, compiled_rules=compiled_rules):
        mi = metadata_from_formats(formats)
        if mi.title is None:
            continue
        ids, dups = db.new_api.add_books([(mi, create_format_map(formats))], add_duplicates=add_duplicates)
        if dups:
            duplicates.append((mi, formats))
            continue
        book_id = next(iter(ids))
        if added_ids is not None:
            added_ids.add(book_id)
        if callable(callback):
            if callback(mi.title):
                break
    return duplicates


def import_book_directory(db, dirpath, callback=None, added_ids=None, compiled_rules=(), add_duplicates=False):
    from calibre.ebooks.metadata.meta import metadata_from_formats

    dirpath = os.path.abspath(dirpath)
    formats = None
    for formats in find_books_in_directory(dirpath, True, compiled_rules=compiled_rules):
        break
    if not formats:
        return
    mi = metadata_from_formats(formats)
    if mi.title is None:
        return
    ids, dups = db.new_api.add_books([(mi, create_format_map(formats))], add_duplicates=add_duplicates)
    if dups:
        return [(mi, formats)]
    book_id = next(iter(ids))
    if added_ids is not None:
        added_ids.add(book_id)
    if callable(callback):
        callback(mi.title)


def recursive_import(db, root, single_book_per_directory=True, callback=None, added_ids=None, compiled_rules=(), add_duplicates=False):
    root = os.path.abspath(root)
    duplicates = []
    for dirpath in os.walk(root):
        func = import_book_directory if single_book_per_directory else import_book_directory_multiple
        res = func(
            db,
            dirpath[0],
            callback=callback,
            added_ids=added_ids,
            compiled_rules=compiled_rules,
            add_duplicates=add_duplicates,
        )
        if res is not None:
            duplicates.extend(res)
        if callable(callback):
            if callback(''):
                break
    return duplicates


def cdb_find_in_dir(dirpath, single_book_per_directory, compiled_rules):
    return find_books_in_directory(
        dirpath,
        single_book_per_directory=single_book_per_directory,
        compiled_rules=compiled_rules,
        listdir_impl=partial(list_only_files_in_dir, sort_by_mtime=True),
    )


def cdb_recursive_find(root, single_book_per_directory=True, compiled_rules=()):
    root = os.path.abspath(root)
    for dirpath in os.walk(root):
        yield from cdb_find_in_dir(dirpath[0], single_book_per_directory, compiled_rules)


def add_catalog(cache, path, title, dbapi=None):
    from calibre.ebooks.metadata.book.base import Metadata
    from calibre.ebooks.metadata.meta import get_metadata
    from calibre.utils.date import utcnow

    fmt = os.path.splitext(path)[1][1:].lower()
    new_book_added = False
    with open(path, 'rb') as stream:
        with cache.write_lock:
            matches = cache._search('title:="{}" and tags:="{}"'.format(title.replace('"', '\\"'), _('Catalog')), None)
            db_id = None
            if matches:
                db_id = list(matches)[0]
            try:
                mi = get_metadata(stream, fmt)
                mi.authors = ['calibre']
            except Exception:
                mi = Metadata(title, ['calibre'])
            mi.title, mi.authors = title, ['calibre']
            mi.author_sort = 'calibre'  # The MOBI/AZW3 format sets author sort to date
            mi.pubdate = mi.timestamp = utcnow()
            if fmt == 'mobi':
                mi.cover, mi.cover_data = None, (None, None)
            if db_id is None:
                mi.tags = [_('Catalog')]
                db_id = cache._create_book_entry(mi, apply_import_tags=False)
                new_book_added = True
            else:
                tags: list[str] = list(cache._field_for('tags', db_id) or ())
                if _('Catalog') not in tags:
                    tags.append(_('Catalog'))
                mi.tags = tags
                cache._set_metadata(db_id, mi)
        cache.add_format(db_id, fmt, stream, dbapi=dbapi)  # Can't keep write lock since post-import hooks might run

    return db_id, new_book_added


def add_news(cache, path, arg, dbapi=None):
    from calibre.ebooks.metadata.meta import get_metadata
    from calibre.utils.date import utcnow

    fmt = os.path.splitext(getattr(path, 'name', path))[1][1:].lower()
    stream = path if hasattr(path, 'read') else open(path, 'rb')
    stream.seek(0)
    mi = get_metadata(stream, fmt, use_libprs_metadata=False, force_read_metadata=True)
    # Force the author to calibre as the auto delete of old news checks for
    # both the author==calibre and the tag News
    mi.authors = ['calibre']
    stream.seek(0)
    with cache.write_lock:
        if mi.series_index is None:
            mi.series_index = cache._get_next_series_num_for(mi.series)
        mi.tags = [_('News')]
        if arg.get('add_title_tag'):
            mi.tags += [arg['title']]
        if arg.get('custom_tags'):
            mi.tags += arg['custom_tags']
        if mi.pubdate is None:
            mi.pubdate = utcnow()
        if mi.timestamp is None:
            mi.timestamp = utcnow()

        db_id = cache._create_book_entry(mi, apply_import_tags=False)
    cache.add_format(db_id, fmt, stream, dbapi=dbapi)  # Can't keep write lock since post-import hooks might run

    if not hasattr(path, 'read'):
        stream.close()
    return db_id



#!/usr/bin/env python
# License: GPLv3 Copyright: 2011, Kovid Goyal <kovid@kovidgoyal.net>


class NoSuchFormat(ValueError):
    pass


class NoTracebackException(Exception):
    suppress_traceback: bool = True


class NoSuchBook(KeyError):
    def __init__(self, book_id):
        KeyError.__init__(self, f'No book with id: {book_id} in database')
        self.book_id = book_id



# License: GPLv3 Copyright: 2008, Marshall T. Vandegrift <llasram@gmail.com>

"""
Microsoft LIT tag and attribute tables.
"""

from calibre.ebooks.lit.maps.html import MAP as HTML_MAP
from calibre.ebooks.lit.maps.opf import MAP as OPF_MAP

OPF_MAP, HTML_MAP



# License: GPLv3 Copyright: 2009, John Schember <john@nachtimwald.com>


def r(*a):
    return list(range(*a))


# Uncommon Characters supported by PML. \\a tag codes
A_CHARS = r(160, 256) + r(130, 136) + r(138, 141) + r(145, 152) + r(153, 157) + [159]

# Extended Unicode characters supported by PML
Latin_ExtendedA = (
    r(0x0100, 0x0104)
    + [
        0x0105,
        0x0107,
        0x010C,
        0x010D,
        0x0112,
        0x0113,
        0x0115,
        0x0117,
        0x0119,
        0x011B,
        0x011D,
        0x011F,
        0x012A,
        0x012B,
        0x012D,
        0x012F,
        0x0131,
        0x0141,
        0x0142,
        0x0144,
        0x0148,
    ]
    + r(0x014B, 0x014E)
    + [0x014F, 0x0151, 0x0155]
    + r(0x0159, 0x015C)
    + [0x015F, 0x0163, 0x0169, 0x016B, 0x016D, 0x0177, 0x017A, 0x017D, 0x017E]
)
Latin_ExtendedB = [
    0x01BF,
    0x01CE,
    0x01D0,
    0x01D2,
    0x01D4,
    0x01E1,
    0x01E3,
    0x01E7,
    0x01EB,
    0x01F0,
    0x0207,
    0x021D,
    0x0227,
    0x022F,
    0x0233,
]
IPA_Extensions = [
    0x0251,
    0x0251,
    0x0254,
    0x0259,
    0x025C,
    0x0265,
    0x026A,
    0x0272,
    0x0283,
    0x0289,
    0x028A,
    0x028C,
    0x028F,
    0x0292,
    0x0294,
    0x029C,
]
Spacing_Modifier_Letters = [0x02BE, 0x02BF, 0x02C7, 0x02C8, 0x02CC, 0x02D0, 0x02D8, 0x02D9]
Greek_and_Coptic = r(0x0391, 0x03A2) + r(0x03A3, 0x03AA) + r(0x03B1, 0x03CA) + [0x03D1, 0x03DD]
Hebrew = r(0x05D0, 0x05EB)
Latin_Extended_Additional = (
    [0x1E0B, 0x1E0D, 0x1E17, 0x1E22, 0x1E24, 0x1E25, 0x1E2B, 0x1E33, 0x1E37, 0x1E41, 0x1E43, 0x1E45, 0x1E47, 0x1E53]
    + r(0x1E59, 0x1E5C)
    + [0x1E61, 0x1E63, 0x1E6B, 0x1E6D, 0x1E6F, 0x1E91, 0x1E93, 0x1E96, 0x1EA1, 0x1ECD, 0x1EF9]
)
General_Punctuation = [0x2011, 0x2038, 0x203D, 0x2042]
Arrows = [0x2190, 0x2192]
Mathematical_Operators = [0x2202, 0x221A, 0x221E, 0x2225, 0x222B, 0x2260, 0x2294, 0x2295, 0x22EE]
Enclosed_Alphanumerics = [0x24CA]
Miscellaneous_Symbols = r(0x261C, 0x2641) + r(0x2642, 0x2648) + r(0x2660, 0x2664) + r(0x266D, 0x2670)
Dingbats = [0x2713, 0x2720]
Private_Use_Area = r(0xE000, 0xE01D) + r(0xE01E, 0xE029) + r(0xE02A, 0xE052)
Alphabetic_Presentation_Forms = [0xFB02, 0xFB2A, 0xFB2B]

# \\U tag codes.
U_CHARS = (
    Latin_ExtendedA
    + Latin_ExtendedB
    + IPA_Extensions
    + Spacing_Modifier_Letters
    + Greek_and_Coptic
    + Hebrew
    + Latin_Extended_Additional
    + General_Punctuation
    + Arrows
    + Mathematical_Operators
    + Enclosed_Alphanumerics
    + Miscellaneous_Symbols
    + Dingbats
    + Private_Use_Area
    + Alphabetic_Presentation_Forms
)


def unipmlcode(char):
    try:
        val = ord(char.encode('cp1252'))
        if val in A_CHARS:
            return f'\\a{val}'
    except Exception:
        pass
    val = ord(char)
    if val in U_CHARS:
        return '\\U%04x'.upper() % val
    else:
        return '?'



#!/usr/bin/env python
# License: GPLv3 Copyright: 2012, Kovid Goyal <kovid@kovidgoyal.net>

FIELDS = [
    'all',
    'title',
    'title_sort',
    'author_sort',
    'authors',
    'comments',
    'cover',
    'formats',
    'id',
    'isbn',
    'library_name',
    'ondevice',
    'pubdate',
    'publisher',
    'rating',
    'series_index',
    'series',
    'size',
    'tags',
    'timestamp',
    'uuid',
    'languages',
    'identifiers',
]

# Allowed fields for template
TEMPLATE_ALLOWED_FIELDS = [
    'author_sort',
    'authors',
    'id',
    'isbn',
    'pubdate',
    'title_sort',
    'publisher',
    'series_index',
    'series',
    'tags',
    'timestamp',
    'title',
    'uuid',
]


class AuthorSortMismatchException(Exception):
    pass


class EmptyCatalogException(Exception):
    pass


class InvalidGenresSourceFieldException(Exception):
    pass



#!/usr/bin/env python
# License: GPLv3 Copyright: 2012, Kovid Goyal <kovid at kovidgoyal.net>

from struct import pack, unpack_from

t1_operand_encoding = [None] * 256
t1_operand_encoding[0:32] = (32) * ['do_operator']
t1_operand_encoding[32:247] = (247 - 32) * ['read_byte']
t1_operand_encoding[247:251] = (251 - 247) * ['read_small_int1']
t1_operand_encoding[251:255] = (255 - 251) * ['read_small_int2']
t1_operand_encoding[255] = 'read_long_int'

t2_operand_encoding = t1_operand_encoding[:]
t2_operand_encoding[28] = 'read_short_int'
t2_operand_encoding[255] = 'read_fixed_1616'

cff_dict_operand_encoding = t2_operand_encoding[:]
cff_dict_operand_encoding[29] = 'read_long_int'
cff_dict_operand_encoding[30] = 'read_real_number'
cff_dict_operand_encoding[255] = 'reserved'

real_nibbles = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '.', 'E', 'E-', None, '-']
real_nibbles_map = {x: i for i, x in enumerate(real_nibbles)}


class ByteCode(dict):
    def read_byte(self, b0, data, index):
        return b0 - 139, index

    def read_small_int1(self, b0, data, index):
        b1 = ord(data[index : index + 1])
        return (b0 - 247) * 256 + b1 + 108, index + 1

    def read_small_int2(self, b0, data, index):
        b1 = ord(data[index : index + 1])
        return -(b0 - 251) * 256 - b1 - 108, index + 1

    def read_short_int(self, b0, data, index):
        (value,) = unpack_from(b'>h', data, index)
        return value, index + 2

    def read_long_int(self, b0, data, index):
        (value,) = unpack_from(b'>l', data, index)
        return value, index + 4

    def read_fixed_1616(self, b0, data, index):
        (value,) = unpack_from(b'>l', data, index)
        return value / 65536.0, index + 4

    def read_real_number(self, b0, data, index):
        number = ''
        while True:
            b = ord(data[index : index + 1])
            index = index + 1
            nibble0 = (b & 0xF0) >> 4
            nibble1 = b & 0x0F
            if nibble0 == 0xF:
                break
            number = number + (real_nibbles[nibble0] or '')
            if nibble1 == 0xF:
                break
            number = number + (real_nibbles[nibble1] or '')
        return float(number), index

    def write_float(self, f, encoding='ignored'):
        s = str(f).upper()
        if s[:2] == '0.':
            s = s[1:]
        elif s[:3] == '-0.':
            s = '-' + s[2:]
        nibbles = []
        while s:
            c = s[0]
            s = s[1:]
            if c == 'E' and s[:1] == '-':
                s = s[1:]
                c = 'E-'
            nibbles.append(real_nibbles_map[c])
        nibbles.append(0xF)
        if len(nibbles) % 2:
            nibbles.append(0xF)
        d = bytearray([30])
        for i in range(0, len(nibbles), 2):
            d.append(nibbles[i] << 4 | nibbles[i + 1])
        return bytes(d)

    def write_int(self, value, encoding='cff'):
        four_byte_op = {'cff': 29, 't1': 255}.get(encoding, None)

        if -107 <= value <= 107:
            code = bytes(bytearray([value + 139]))
        elif 108 <= value <= 1131:
            value = value - 108
            code = bytes(bytearray([(value >> 8) + 247, (value & 0xFF)]))
        elif -1131 <= value <= -108:
            value = -value - 108
            code = bytes(bytearray([(value >> 8) + 251, (value & 0xFF)]))
        elif four_byte_op is None:
            # T2 only supports 2 byte ints
            code = bytes(bytearray([28])) + pack(b'>h', value)
        else:
            code = bytes(bytearray([four_byte_op])) + pack(b'>l', value)
        return code

    def write_offset(self, value):
        return bytes(bytearray([29])) + pack(b'>l', value)

    def write_number(self, value, encoding='cff'):
        f = self.write_float if isinstance(value, float) else self.write_int
        return f(value, encoding)


class Dict(ByteCode):
    operand_encoding = cff_dict_operand_encoding
    TABLE = ()
    FILTERED = frozenset()
    OFFSETS = frozenset()

    def __init__(self):
        ByteCode.__init__(self)

        self.operators = {op: (name, arg) for op, name, arg, default in self.TABLE}
        self.defaults = {name: default for op, name, arg, default in self.TABLE}

    def safe_get(self, name):
        return self.get(name, self.defaults[name])

    def decompile(self, strings, global_subrs, data):
        self.strings = strings
        self.global_subrs = global_subrs
        self.stack = []
        index = 0
        while index < len(data):
            b0 = ord(data[index : index + 1])
            index += 1
            handler = getattr(self, str(self.operand_encoding[b0]))
            value, index = handler(b0, data, index)
            if value is not None:
                self.stack.append(value)

    def do_operator(self, b0, data, index):
        if b0 == 12:
            op = (b0, ord(data[index : index + 1]))
            index += 1
        else:
            op = b0
        operator, arg_type = self.operators[op]
        self.handle_operator(operator, arg_type)
        return None, index

    def handle_operator(self, operator, arg_type):
        if isinstance(arg_type, tuple):
            value = ()
            for i in range(len(arg_type) - 1, -1, -1):
                arg = arg_type[i]
                arghandler = getattr(self, 'arg_' + arg)
                value = (arghandler(operator),) + value
        else:
            arghandler = getattr(self, 'arg_' + arg_type)
            value = arghandler(operator)
        self[operator] = value

    def arg_number(self, name):
        return self.stack.pop()

    def arg_SID(self, name):
        return self.strings[self.stack.pop()]

    def arg_array(self, name):
        ans = self.stack[:]
        del self.stack[:]
        return ans

    def arg_delta(self, name):
        out = []
        current = 0
        for v in self.stack:
            current = current + v
            out.append(current)
        del self.stack[:]
        return out

    def compile(self, strings):
        data = []
        for op, name, arg, default in self.TABLE:
            if name in self.FILTERED:
                continue
            val = self.safe_get(name)
            opcode = bytes(bytearray(op if isinstance(op, tuple) else [op]))
            if val != self.defaults[name]:
                self.encoding_offset = name in self.OFFSETS
                if isinstance(arg, tuple):
                    if len(val) != len(arg):
                        raise ValueError(f'Invalid argument {val} for operator: {op}')
                    for typ, v in zip(arg, val):
                        if typ == 'SID':
                            val = strings(val)
                        data.append(getattr(self, 'encode_' + str(typ))(v))
                else:
                    if arg == 'SID':
                        val = strings(val)
                    data.append(getattr(self, 'encode_' + arg)(val))
                data.append(opcode)
        self.raw = b''.join(data)
        return self.raw

    def encode_number(self, val):
        if self.encoding_offset:
            return self.write_offset(val)
        return self.write_number(val)

    def encode_SID(self, val):
        return self.write_int(val)

    def encode_array(self, val):
        return b''.join(map(self.encode_number, val))

    def encode_delta(self, value):
        out = []
        last = 0
        for v in value:
            out.append(v - last)
            last = v
        return self.encode_array(out)


class TopDict(Dict):
    TABLE = (
        # opcode     name                  argument type   default
        (
            (12, 30),
            'ROS',
            ('SID', 'SID', 'number'),
            None,
        ),
        (
            (12, 20),
            'SyntheticBase',
            'number',
            None,
        ),
        (
            0,
            'version',
            'SID',
            None,
        ),
        (
            1,
            'Notice',
            'SID',
            None,
        ),
        (
            (12, 0),
            'Copyright',
            'SID',
            None,
        ),
        (
            2,
            'FullName',
            'SID',
            None,
        ),
        (
            (12, 38),
            'FontName',
            'SID',
            None,
        ),
        (
            3,
            'FamilyName',
            'SID',
            None,
        ),
        (
            4,
            'Weight',
            'SID',
            None,
        ),
        (
            (12, 1),
            'isFixedPitch',
            'number',
            0,
        ),
        (
            (12, 2),
            'ItalicAngle',
            'number',
            0,
        ),
        (
            (12, 3),
            'UnderlinePosition',
            'number',
            None,
        ),
        (
            (12, 4),
            'UnderlineThickness',
            'number',
            50,
        ),
        (
            (12, 5),
            'PaintType',
            'number',
            0,
        ),
        (
            (12, 6),
            'CharstringType',
            'number',
            2,
        ),
        (
            (12, 7),
            'FontMatrix',
            'array',
            [0.001, 0, 0, 0.001, 0, 0],
        ),
        (
            13,
            'UniqueID',
            'number',
            None,
        ),
        (
            5,
            'FontBBox',
            'array',
            [0, 0, 0, 0],
        ),
        (
            (12, 8),
            'StrokeWidth',
            'number',
            0,
        ),
        (
            14,
            'XUID',
            'array',
            None,
        ),
        (
            (12, 21),
            'PostScript',
            'SID',
            None,
        ),
        (
            (12, 22),
            'BaseFontName',
            'SID',
            None,
        ),
        (
            (12, 23),
            'BaseFontBlend',
            'delta',
            None,
        ),
        (
            (12, 31),
            'CIDFontVersion',
            'number',
            0,
        ),
        (
            (12, 32),
            'CIDFontRevision',
            'number',
            0,
        ),
        (
            (12, 33),
            'CIDFontType',
            'number',
            0,
        ),
        (
            (12, 34),
            'CIDCount',
            'number',
            8720,
        ),
        (
            15,
            'charset',
            'number',
            0,
        ),
        (
            (12, 35),
            'UIDBase',
            'number',
            None,
        ),
        (
            16,
            'Encoding',
            'number',
            0,
        ),
        (
            18,
            'Private',
            ('number', 'number'),
            None,
        ),
        (
            (12, 37),
            'FDSelect',
            'number',
            None,
        ),
        (
            (12, 36),
            'FDArray',
            'number',
            None,
        ),
        (
            17,
            'CharStrings',
            'number',
            None,
        ),
    )

    # We will not write these operators out
    FILTERED = {
        'ROS',
        'SyntheticBase',
        'UniqueID',
        'XUID',
        'CIDFontVersion',
        'CIDFontRevision',
        'CIDFontType',
        'CIDCount',
        'UIDBase',
        'Encoding',
        'FDSelect',
        'FDArray',
    }
    OFFSETS = {'charset', 'Encoding', 'CharStrings', 'Private'}


class PrivateDict(Dict):
    TABLE = (
        #   opcode     name                  argument type   default
        (
            6,
            'BlueValues',
            'delta',
            None,
        ),
        (
            7,
            'OtherBlues',
            'delta',
            None,
        ),
        (
            8,
            'FamilyBlues',
            'delta',
            None,
        ),
        (
            9,
            'FamilyOtherBlues',
            'delta',
            None,
        ),
        (
            (12, 9),
            'BlueScale',
            'number',
            0.039625,
        ),
        (
            (12, 10),
            'BlueShift',
            'number',
            7,
        ),
        (
            (12, 11),
            'BlueFuzz',
            'number',
            1,
        ),
        (
            10,
            'StdHW',
            'number',
            None,
        ),
        (
            11,
            'StdVW',
            'number',
            None,
        ),
        (
            (12, 12),
            'StemSnapH',
            'delta',
            None,
        ),
        (
            (12, 13),
            'StemSnapV',
            'delta',
            None,
        ),
        (
            (12, 14),
            'ForceBold',
            'number',
            0,
        ),
        (
            (12, 15),
            'ForceBoldThreshold',
            'number',
            None,
        ),  # deprecated
        (
            (12, 16),
            'lenIV',
            'number',
            None,
        ),  # deprecated
        (
            (12, 17),
            'LanguageGroup',
            'number',
            0,
        ),
        (
            (12, 18),
            'ExpansionFactor',
            'number',
            0.06,
        ),
        (
            (12, 19),
            'initialRandomSeed',
            'number',
            0,
        ),
        (
            20,
            'defaultWidthX',
            'number',
            0,
        ),
        (
            21,
            'nominalWidthX',
            'number',
            0,
        ),
        (
            19,
            'Subrs',
            'number',
            None,
        ),
    )

    OFFSETS = {'Subrs'}



#!/usr/bin/env python
# License: GPLv3 Copyright: 2010, Kovid Goyal <kovid@kovidgoyal.net>

from functools import partial

from qt.core import QIcon, Qt

from calibre.constants import DEBUG, ismacos
from calibre.customize.ui import preferences_plugins
from calibre.gui2 import error_dialog, show_restart_warning
from calibre.gui2.actions import InterfaceAction
from calibre.gui2.preferences.main import Preferences
from calibre.utils.localization import _


class PreferencesAction(InterfaceAction):
    name = 'Preferences'
    action_spec = (_('Preferences'), 'config.png', _('Configure calibre'), 'Ctrl+P')
    action_add_menu = True
    action_menu_clone_qaction = _('Change calibre behavior')

    def genesis(self):
        pm = self.qaction.menu()
        assert pm is not None
        cm = partial(self.create_menu_action, pm)
        if ismacos:
            pm.addAction(QIcon.ic('config.png'), _('Preferences'), self.do_config)
        cm('welcome wizard', _('Run Welcome wizard'), icon='wizard.png', triggered=self.gui.run_wizard)
        cm(
            'plugin updater',
            _('Get plugins to enhance calibre'),
            icon='plugins/plugin_updater.png',
            triggered=self.get_plugins,
        )
        pm.addSeparator()
        if not DEBUG:
            cm(
                'restart',
                _('Restart in debug mode'),
                icon='debug.png',
                triggered=self.debug_restart,
                shortcut='Ctrl+Shift+R',
            )
        cm(
            'restart_without_plugins',
            _('Restart ignoring third party plugins'),
            icon='debug.png',
            triggered=self.no_plugins_restart,
            shortcut='Ctrl+Alt+Shift+R',
        )

        self.preferences_menu = pm
        for x in (self.gui.preferences_action, self.qaction):
            x.triggered.connect(self.do_config)
        self.action_map = {}

    def initialization_complete(self):
        # Add the individual preferences to the menu.
        # First, sort them into the same order as shown in the preferences dialog
        plugins = sorted(preferences_plugins(), key=lambda p: p.category_order * 100 + p.name_order)

        pm = self.preferences_menu
        assert pm is not None
        # The space pushes the separator a bit away from the text
        pm.addSection(_('Sections') + ' ')

        config_icon = QIcon.ic('config.png')
        current_cat = 0
        for p in plugins:
            if p.category_order != current_cat:
                current_cat = p.category_order
                cm = pm.addMenu(p.gui_category.replace('&', '&&'))
                cm.setIcon(config_icon)
            self.action_map[p.name] = self.create_menu_action(
                cm,
                p.name,
                p.gui_name.replace('&', '&&'),
                icon=QIcon.ic(p.icon),
                shortcut=None,
                shortcut_name=p.gui_name,
                triggered=partial(self.do_config, initial_plugin=(p.category, p.name), close_after_initial=True),
            )

    def get_plugins(self):
        from calibre.gui2.dialogs.plugin_updater import FILTER_NOT_INSTALLED, PluginUpdaterDialog

        d = PluginUpdaterDialog(self.gui, initial_filter=FILTER_NOT_INSTALLED)
        d.exec()
        if d.do_restart:
            self.gui.quit(restart=True)

    def do_config(self, checked=False, initial_plugin=None, close_after_initial=False):
        if self.gui.job_manager.has_jobs():
            d = error_dialog(self.gui, _('Cannot configure'), _('Cannot configure while there are running jobs.'))
            d.exec()
            return
        if self.gui.must_restart_before_config:
            do_restart = show_restart_warning(_('Cannot configure before calibre is restarted.'))
            if do_restart:
                self.gui.quit(restart=True)
            return
        d = Preferences(self.gui, initial_plugin=initial_plugin, close_after_initial=close_after_initial)
        d.run_wizard_requested.connect(self.gui.run_wizard, type=Qt.ConnectionType.QueuedConnection)
        d.exec()
        if d.do_restart:
            self.gui.quit(restart=True)

    def debug_restart(self, *args):
        self.gui.quit(restart=True, debug_on_restart=True)

    def no_plugins_restart(self, *args):
        self.gui.quit(restart=True, no_plugins_on_restart=True)



#!/usr/bin/env python
# License: GPLv3 Copyright: 2009, Kovid Goyal <kovid@kovidgoyal.net>

import re

from calibre.devices.usbms.driver import USBMS
from calibre.utils.localization import _


class IRIVER_STORY(USBMS):
    name = 'Iriver Story Device Interface'
    gui_name = 'Iriver Story'
    description = _('Communicate with the Iriver Story reader.')
    author = 'Kovid Goyal'
    supported_platforms = ['windows', 'osx', 'linux']

    # Ordered list of supported formats
    FORMATS = ['epub', 'fb2', 'pdf', 'djvu', 'txt']

    VENDOR_ID = [0x1006]
    PRODUCT_ID = [0x4023, 0x4024, 0x4025, 0x4034, 0x4037]
    BCD = [0x0323, 0x0326, 0x226]

    VENDOR_NAME = 'IRIVER'
    WINDOWS_MAIN_MEM = ['STORY', 'STORY_EB05', 'STORY_WI-FI', 'STORY_EB07', 'STORY_EB12']
    WINDOWS_MAIN_MEM = re.compile(r'({})&'.format('|'.join(WINDOWS_MAIN_MEM)))
    WINDOWS_CARD_A_MEM = ['STORY', 'STORY_SD', 'STORY_EB12_SD']
    WINDOWS_CARD_A_MEM = re.compile(r'({})&'.format('|'.join(WINDOWS_CARD_A_MEM)))

    # OSX_MAIN_MEM = 'Kindle Internal Storage Media'
    # OSX_CARD_A_MEM = 'Kindle Card Storage Media'

    MAIN_MEMORY_VOLUME_LABEL = 'Story Main Memory'
    STORAGE_CARD_VOLUME_LABEL = 'Story Storage Card'
    EBOOK_DIR_MAIN = 'Book'

    SUPPORTS_SUB_DIRS = True



# License: GPLv3 Copyright: 2011, John Schember <john@nachtimwald.com>

"""
Config widget access functions for configuring the store action.
"""


def config_widget():
    from calibre.gui2.store.config.search.search_widget import StoreConfigWidget

    return StoreConfigWidget()


def save_settings(config_widget):
    config_widget.save_settings()



#!/usr/bin/env python
# License: GPLv3 Copyright: 2010, Kovid Goyal <kovid@kovidgoyal.net>

from qt.core import QListWidgetItem, QMenu, QPoint, Qt, QTabWidget

from calibre.gui2.preferences import ConfigWidgetBase, test_widget
from calibre.gui2.preferences.look_feel_ui import Ui_Form
from calibre.utils.localization import _


class ConfigWidget(ConfigWidgetBase, Ui_Form):
    def genesis(self, gui):
        self.gui = gui

        self.tabWidget.setCurrentIndex(0)
        if tab_bar := self.tabWidget.tabBar():
            tab_bar.setVisible(False)

        for i in range(self.tabWidget.count()):
            self.sections_view.addItem(QListWidgetItem(self.tabWidget.tabIcon(i), self.tabWidget.tabText(i).replace('&', '')))
        self.sections_view.setCurrentRow(self.tabWidget.currentIndex())
        self.sections_view.currentRowChanged.connect(self.tabWidget.setCurrentIndex)
        self.sections_view.setMaximumWidth(self.sections_view.sizeHintForColumn(0) + 16)
        self.sections_view.setSpacing(4)
        self.sections_view.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.sections_view.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.sections_view.customContextMenuRequested.connect(self.show_context_menu_for_section)

        self.tabWidget.currentWidget().setFocus(Qt.FocusReason.OtherFocusReason)

    def initial_tab_changed(self):
        self.sections_view.setCurrentRow(self.tabWidget.currentIndex())

    def _restore_widget_defaults(self, widget):
        if hasattr(widget, 'restore_defaults'):
            widget.restore_defaults()
        elif isinstance(widget, QTabWidget):
            for i in range(widget.count()):
                sw = widget.widget(i)
                if sw is not None and hasattr(sw, 'restore_defaults'):
                    sw.restore_defaults()

    def restore_defaults(self, *args):
        ConfigWidgetBase.restore_defaults(self)
        for w in self.tabWidget.all_widgets:
            self._restore_widget_defaults(w)
        self.changed_signal.emit()

    def refresh_gui(self, gui):
        m = gui.library_view.model()
        m.update_db_prefs_cache()
        m.beginResetModel(), m.endResetModel()
        gui.tags_view.model().reset_tag_browser()

    def show_context_menu_for_section(self, pos: QPoint) -> None:
        if item := self.sections_view.itemAt(pos):
            menu = QMenu(self.sections_view)
            menu.addAction(_('Restore defaults for {}').format(item.data(Qt.ItemDataRole.DisplayRole)))
            num = self.sections_view.indexFromItem(item).row()
            widget = tuple(self.tabWidget.all_widgets)[num]
            pos = self.sections_view.mapToGlobal(pos)
            can_restore = hasattr(widget, 'restore_defaults') or isinstance(widget, QTabWidget)
            if can_restore and menu.exec(pos):  # type: ignore
                self._restore_widget_defaults(widget)
                self.changed_signal.emit()


if __name__ == '__main__':
    from calibre.gui2 import Application

    app = Application([])
    test_widget('Interface', 'Look & Feel')
